"""
Library to handle Dynet networks.

@ Author      : Troy Kelly
@ Date        : 23 Sept 2018
@ Description : Philips Dynalite Library - Unofficial interface for Philips Dynalite over RS485

@ Notes:        Requires a RS485 to IP gateway (Do not use the Dynalite one - use something cheaper)
"""

import json
from typing import List, Optional

from .opcodes import OpcodeType, SyncType


class PacketError(Exception):
    """Class for Dynet packet errors."""

    def __init__(self, message: str) -> None:
        """Initialize the error."""
        self.message = message
        super().__init__(message)


class DynetPacket:
    """Class for a Dynet network packet."""

    def __init__(
        self,
        msg: List[int] = None,
        area: int = -1,
        command: int = -1,
        data: List[int] = None,
    ) -> None:
        """Initialize the packet."""
        # Either msg is defined or area/command/data, but not both
        if msg is None:
            # Can only have one of the two init options
            assert area != -1 and command != -1 and data
            self.area = area
            self.command = command
            self.data = data
            self._msg: List[int] = [
                SyncType.LOGICAL.value,
                area,
                data[0],
                command,
                data[1],
                data[2],
                255,  # join
            ]
            self._msg.append(self.calc_sum(self._msg))
        else:
            # Can only have one of the two init options
            assert area == -1 and command == -1 and not data
            if len(msg) != 8:
                raise PacketError(f"Wrong message size of {len(msg)} - should be 8")
            self._msg = msg
            sync = msg[0]
            assert sync == SyncType.LOGICAL.value
            self.area = msg[1]
            self.command = msg[3]
            self.data = [msg[2], msg[4], msg[5]]
            chk = msg[7]
            if chk != self.calc_sum(msg):
                raise PacketError(
                    f"Message with the wrong checksum - {[int(byte) for byte in msg]}"
                )

    @property
    def opcode_type(self) -> Optional[str]:
        """Return the alphabetic representation of the opcode if known or None."""
        if OpcodeType.has_value(self.command):
            return OpcodeType(self.command).name
        return None

    @property
    def raw_msg(self) -> List[int]:
        """Return the raw message as a list of integers."""
        return self._msg

    @property
    def msg(self) -> bytearray:
        """Get the byte array for the message to send."""
        return bytearray(self._msg)

    @staticmethod
    def calc_sum(msg: List[int]) -> int:
        """Calculate the checksum."""
        msg = msg[:7]
        return -(sum(msg) % 256) & 0xFF

    def __repr__(self):
        """Print the packet."""
        return json.dumps(self.__dict__)

    @staticmethod
    def set_channel_level_packet(
        area: int, channel: int, level: float, fade: float
    ) -> "DynetPacket":
        """Create a packet to set level of a channel."""
        channel_bank = 0xFF if (channel <= 4) else (int((channel - 1) / 4) - 1)
        target_level = int(255 - 254 * level)
        opcode = 0x80 + ((channel - 1) % 4)
        fade_time = int(fade / 0.02)
        if (fade_time) > 0xFF:
            fade_time = 0xFF
        return DynetPacket(
            area=area, command=opcode, data=[target_level, channel_bank, fade_time]
        )

    @staticmethod
    def select_area_preset_packet(area: int, preset: int, fade: float) -> "DynetPacket":
        """Create a packet to select a preset in an area."""
        preset = preset - 1
        bank = int((preset) / 8)
        opcode = preset - (bank * 8)
        if opcode > 3:
            opcode = opcode + 6
        fade_low = int(fade / 0.02) - (int((fade / 0.02) / 256) * 256)
        fade_high = int((fade / 0.02) / 256)
        return DynetPacket(area=area, command=opcode, data=[fade_low, fade_high, bank])

    @staticmethod
    def request_channel_level_packet(area: int, channel: int) -> "DynetPacket":
        """Create a packet to request the level of a specific channel."""
        return DynetPacket(
            area=area,
            command=OpcodeType.REQUEST_CHANNEL_LEVEL.value,
            data=[channel - 1, 0, 0],
        )

    @staticmethod
    def stop_channel_fade_packet(area: int, channel: int) -> "DynetPacket":
        """Create a packet to stop fade of a channel."""
        return DynetPacket(
            area=area, command=OpcodeType.STOP_FADING.value, data=[channel - 1, 0, 0]
        )

    @staticmethod
    def request_area_preset_packet(area: int, query_channel: int) -> "DynetPacket":
        """Create a packet to request the current preset in an area."""
        return DynetPacket(
            area=area,
            command=OpcodeType.REQUEST_PRESET.value,
            data=[0, query_channel - 1, 0],
        )

    @staticmethod
    def report_channel_level_packet(
        area: int, channel: int, target_level: float, actual_level: float
    ) -> "DynetPacket":
        """Create a packet that reports the level of a channel."""
        assert 0.0 <= target_level <= 1.0
        assert 0.0 <= actual_level <= 1.0
        return DynetPacket(
            area=area,
            command=OpcodeType.REPORT_CHANNEL_LEVEL.value,
            data=[
                channel - 1,
                int(255 - 254 * target_level),
                int(255 - 254 * actual_level),
            ],
        )

    @staticmethod
    def report_area_preset_packet(area: int, preset: int) -> "DynetPacket":
        """Create a packet that reports the current preset in an area."""
        return DynetPacket(
            area=area, command=OpcodeType.REPORT_PRESET.value, data=[preset - 1, 0, 0]
        )

    @staticmethod
    def fade_area_channel_preset_packet(
        area: int, channel: int, preset: int, fade: float
    ) -> "DynetPacket":
        """Create a packet that reports the current preset in an area."""
        fade_time = int(fade / 0.02)
        if (fade_time) > 0xFF:
            fade_time = 0xFF
        return DynetPacket(
            area=area,
            command=OpcodeType.FADE_CHANNEL_AREA_TO_PRESET.value,
            data=[channel - 1, preset - 1, fade_time],
        )



"""Support for the Dynalite devices."""
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Callable, Dict, List

from .const import CONF_CHANNEL

if TYPE_CHECKING:  # pragma: no cover
    from .dynalite_devices import DynaliteDevices


class DynaliteBaseDevice(ABC):
    """Base class for Dynalite devices."""

    def __init__(self, area: int, bridge: "DynaliteDevices", hidden: bool) -> None:
        """Initialize the device."""
        self._area = area
        self._bridge = bridge
        self._listeners: List[Callable[[DynaliteBaseDevice, bool], None]] = []
        self._hidden = hidden

    @property
    def area_name(self) -> str:
        """Return the name of the area."""
        return self._bridge.get_area_name(self._area)

    @property
    def get_master_area(self) -> str:
        """Get the master area when combining entities from different Dynet areas to the same area."""
        return self._bridge.get_master_area(self._area)

    @property
    def hidden(self) -> bool:
        """Get whether the device should be hidden from the calling platform."""
        return self._hidden

    def add_listener(
        self, listener: Callable[["DynaliteBaseDevice", bool], None]
    ) -> None:
        """Add a listener for changes."""
        self._listeners.append(listener)

    def update_listeners(self, stop_fade: bool = False) -> None:
        """Update all listeners."""
        for listener in self._listeners:
            listener(self, stop_fade)

    @abstractmethod
    def init_level(self, level: float):
        """Initialize the level."""


class DynaliteChannelBaseDevice(DynaliteBaseDevice):
    """Representation of a Dynalite Channel as a Home Assistant device."""

    def __init__(
        self, area: int, channel: int, bridge: "DynaliteDevices", hidden: bool
    ) -> None:
        """Initialize the device."""
        self._channel = channel
        super().__init__(area, bridge, hidden)

    @property
    def available(self) -> bool:
        """Return if device is available."""
        return self._bridge.available(CONF_CHANNEL, self._area, self._channel)

    @property
    def name(self) -> str:
        """Return the name of the device."""
        return self._bridge.get_channel_name(self._area, self._channel)

    @property
    def unique_id(self) -> str:
        """Return the ID of this device."""
        return "dynalite_area_" + str(self._area) + "_channel_" + str(self._channel)

    def stop_fade(self) -> None:
        """Update the listeners if STOP FADE is received."""
        self.update_listeners(True)


class DynaliteMultiDevice(DynaliteBaseDevice):
    """Representation of two Dynalite Presets as an on/off switch."""

    def __init__(
        self, num_devices: int, area: int, bridge: "DynaliteDevices", hidden: bool
    ) -> None:
        """Initialize the device."""
        self._devices: Dict[int, DynaliteBaseDevice] = {}
        super().__init__(area, bridge, hidden)

    @property
    def name(self) -> str:
        """Return the name of the device."""
        return self._bridge.get_multi_name(self._area)

    def get_device(self, devnum: int) -> DynaliteBaseDevice:
        """Get one of the devices."""
        assert devnum in self._devices
        return self._devices[devnum]

    def set_device(self, devnum: int, device: DynaliteBaseDevice) -> None:
        """Set one of the attached devices."""
        self._devices[devnum] = device
        device.add_listener(self.listener)

    def listener(self, device: DynaliteBaseDevice, stop_fade: bool) -> None:
        """Update the device since its internal devices changed."""
        # pylint: disable=unused-argument
        self._bridge.update_device(self)



"""Constants for the Dynalite component."""
import logging
from typing import Dict, Union

LOGGER = logging.getLogger(__package__)
DOMAIN = "dynalite-devices"
DATA_CONFIGS = "dynalite_configs"

CONF_ACT_LEVEL = "actual_level"
CONF_ACTION = "action"
CONF_ACTION_CMD = "cmd"
CONF_ACTION_REPORT = "report"
CONF_ACTION_STOP = "stop"
CONF_ACTION_PRESET = "preset"
CONF_ACTIVE = "active"
ACTIVE_INIT = "init"
ACTIVE_OFF = "off"
ACTIVE_ON = "on"
CONF_AREA = "area"
CONF_AREA_OVERRIDE = "areaoverride"
CONF_AUTO_DISCOVER = "autodiscover"
CONF_CHANNEL = "channel"
CONF_CHANNEL_COVER = "channelcover"
CONF_CHANNEL_TYPE = "type"
CONF_CLOSE_PRESET = "close"
CONF_DEFAULT = "default"
CONF_DEVICE_CLASS = "class"
CONF_DURATION = "duration"
CONF_FADE = "fade"
CONF_HIDDEN_ENTITY = "hidden"
CONF_HOST = "host"
CONF_LEVEL = "level"
CONF_NAME = "name"
CONF_NO_DEFAULT = "nodefault"
CONF_NONE = "none"
CONF_OPEN_PRESET = "open"
CONF_POLL_TIMER = "polltimer"
CONF_PORT = "port"
CONF_PRESET = "preset"
CONF_QUERY_CHANNEL = "query_channel"
CONF_ROOM = "room"
CONF_ROOM_OFF = "room_off"
CONF_ROOM_ON = "room_on"
CONF_STOP_PRESET = "stop"
CONF_TEMPLATE = "template"
CONF_TILT_TIME = "tilt"
CONF_TIME_COVER = "timecover"
CONF_TRGT_LEVEL = "target_level"
CONF_TRIGGER = "trigger"

ATTR_BRIGHTNESS = "brightness"
ATTR_POSITION = "position"
ATTR_TILT_POSITION = "tilt_position"

DEFAULT_CHANNEL_TYPE = "light"
DEFAULT_COVER_CLASS = "shutter"
DEFAULT_NAME = "dynalite"
DEFAULT_PORT = 12345
DEFAULT_QUERY_CHANNEL = 1
DEFAULT_TEMPLATES: Dict[str, Dict[str, Union[str, int]]] = {
    CONF_ROOM: {CONF_ROOM_ON: "1", CONF_ROOM_OFF: "4"},
    CONF_TRIGGER: {CONF_TRIGGER: "1"},
    CONF_HIDDEN_ENTITY: {},
    CONF_TIME_COVER: {
        CONF_CHANNEL_COVER: "0",
        CONF_DEVICE_CLASS: DEFAULT_COVER_CLASS,
        CONF_OPEN_PRESET: "1",
        CONF_CLOSE_PRESET: "2",
        CONF_STOP_PRESET: "4",
        CONF_DURATION: 60,
        CONF_TILT_TIME: 0,
    },
}
DEFAULT_PRESETS = {
    "1": {CONF_NAME: "On", CONF_LEVEL: 1.0},
    "4": {CONF_NAME: "Off", CONF_LEVEL: 0.0},
}

EVENT_CHANNEL = "CHANNEL"
EVENT_CONNECTED = "CONNECTED"
EVENT_DISCONNECTED = "DISCONNECTED"
EVENT_PRESET = "PRESET"
EVENT_PACKET = "PACKET"

NOTIFICATION_PACKET = "PACKET"
NOTIFICATION_PRESET = "PRESET"

CONNECTION_RETRY_DELAY = 1  # seconds to reconnect
MESSAGE_DELAY = 0.2  # seconds between sending



"""Tests for Dynalite switches."""
import asyncio

import pytest

import dynalite_devices_lib.const as dyn_const
from dynalite_devices_lib.dynet import DynetPacket

from .common import packet_notification, preset_notification


@pytest.mark.asyncio
async def test_preset_switch(mock_gateway):
    """Test a Dynalite preset."""
    name = "NAME"
    preset_name = "PRESET"
    [device1, device4] = mock_gateway.configure_dyn_dev(
        {
            dyn_const.CONF_ACTIVE: False,
            dyn_const.CONF_AREA: {
                "1": {
                    dyn_const.CONF_NAME: name,
                    dyn_const.CONF_NO_DEFAULT: True,
                    dyn_const.CONF_PRESET: {
                        "1": {
                            dyn_const.CONF_NAME: preset_name,
                            dyn_const.CONF_FADE: 0.5,
                        },
                        "4": {dyn_const.CONF_FADE: 0.7},
                    },
                }
            },
        },
        2,
    )
    assert await mock_gateway.async_setup_dyn_dev()
    await mock_gateway.check_single_update(None)
    assert device1.category == "switch"
    assert device4.category == "switch"
    assert device1.name == f"{name} {preset_name}"
    assert device4.name == f"{name} Preset 4"
    assert device1.unique_id == "dynalite_area_1_preset_1"
    await device1.async_turn_on()
    await mock_gateway.check_single_write(
        DynetPacket.select_area_preset_packet(1, 1, 0.5)
    )
    await mock_gateway.check_updates([device1, device4])
    await mock_gateway.check_notifications([preset_notification(1, 1)])
    assert device1.is_on
    assert not device4.is_on
    await device4.async_turn_on()
    await asyncio.sleep(0.1)
    await mock_gateway.check_single_write(
        DynetPacket.select_area_preset_packet(1, 4, 0.7)
    )
    await mock_gateway.check_updates([device1, device4])
    await mock_gateway.check_notifications([preset_notification(1, 4)])
    assert device4.is_on
    assert not device1.is_on
    await device4.async_turn_off()
    await asyncio.sleep(0.1)
    await mock_gateway.check_writes([])
    assert not device4.is_on
    assert not device1.is_on
    packet_to_send = DynetPacket.select_area_preset_packet(1, 1, 0.2)
    await mock_gateway.receive(packet_to_send)
    await mock_gateway.check_updates([device1, device4])
    await mock_gateway.check_notifications(
        [packet_notification(packet_to_send.raw_msg), preset_notification(1, 1)]
    )
    assert not device4.is_on
    assert device1.is_on
    packet_to_send = DynetPacket.report_area_preset_packet(1, 4)
    await mock_gateway.receive(packet_to_send)
    await mock_gateway.check_updates([device1, device4])
    await mock_gateway.check_notifications(
        [packet_notification(packet_to_send.raw_msg), preset_notification(1, 4)]
    )
    assert device4.is_on
    assert not device1.is_on
    device1.init_level(2)
    assert device1.is_on
    device1.init_level(0)
    assert not device1.is_on


@pytest.mark.asyncio
async def test_channel_switch(mock_gateway):
    """Test a Dynalite channel that is a switch."""
    name = "NAME"
    [device] = mock_gateway.configure_dyn_dev(
        {
            dyn_const.CONF_ACTIVE: False,
            dyn_const.CONF_AREA: {
                "1": {
                    dyn_const.CONF_NAME: name,
                    dyn_const.CONF_NO_DEFAULT: True,
                    dyn_const.CONF_CHANNEL: {
                        "1": {
                            dyn_const.CONF_FADE: 0.5,
                            dyn_const.CONF_CHANNEL_TYPE: "switch",
                        }
                    },
                }
            },
        }
    )
    assert await mock_gateway.async_setup_dyn_dev()
    await mock_gateway.check_single_update(None)
    assert device.category == "switch"
    assert device.name == f"{name} Channel 1"
    assert device.unique_id == "dynalite_area_1_channel_1"
    assert device.available
    assert device.area_name == name
    assert device.get_master_area == name
    await device.async_turn_on()
    await mock_gateway.check_single_update(device)
    await mock_gateway.check_single_write(
        DynetPacket.set_channel_level_packet(1, 1, 1.0, 0.5)
    )
    assert device.is_on
    await device.async_turn_off()
    await mock_gateway.check_single_update(device)
    await mock_gateway.check_single_write(
        DynetPacket.set_channel_level_packet(1, 1, 0.0, 0.5)
    )
    assert not device.is_on
    device.init_level(2)
    assert device.is_on
    device.init_level(0)
    assert not device.is_on
    with pytest.raises(ValueError):
        device.init_level(-1)


@pytest.mark.asyncio
async def test_room_switch(mock_gateway):
    """Test a room switch with two presets."""
    name = "NAME"
    [on_device, off_device, room_device] = mock_gateway.configure_dyn_dev(
        {
            dyn_const.CONF_ACTIVE: False,
            dyn_const.CONF_AREA: {
                "1": {
                    dyn_const.CONF_NAME: name,
                    dyn_const.CONF_TEMPLATE: dyn_const.CONF_ROOM,
                    dyn_const.CONF_PRESET: {"1": {}, "4": {}},
                }
            },
        },
        3,
    )
    assert await mock_gateway.async_setup_dyn_dev()
    await mock_gateway.check_single_update(None)
    for device in [on_device, off_device, room_device]:
        assert device.category == "switch"
    assert room_device.name == name
    assert device.unique_id == "dynalite_area_1_room_switch"
    assert device.available
    await room_device.async_turn_on()
    await mock_gateway.check_updates([on_device, off_device, room_device])
    await mock_gateway.check_notifications([preset_notification(1, 1)])
    await mock_gateway.check_single_write(
        DynetPacket.select_area_preset_packet(1, 1, 0)
    )
    assert room_device.is_on
    assert on_device.is_on
    assert not off_device.is_on
    await room_device.async_turn_off()
    await mock_gateway.check_updates([on_device, off_device, room_device], True)
    await mock_gateway.check_notifications([preset_notification(1, 4)])
    await mock_gateway.check_single_write(
        DynetPacket.select_area_preset_packet(1, 4, 0)
    )
    assert not room_device.is_on
    assert not on_device.is_on
    assert off_device.is_on
    await on_device.async_turn_on()
    await mock_gateway.check_updates([on_device, off_device, room_device], True)
    await mock_gateway.check_notifications([preset_notification(1, 1)])
    await mock_gateway.check_single_write(
        DynetPacket.select_area_preset_packet(1, 1, 0)
    )
    assert room_device.is_on
    assert on_device.is_on
    assert not off_device.is_on
    await off_device.async_turn_on()
    await mock_gateway.check_updates([on_device, off_device, room_device], True)
    await mock_gateway.check_notifications([preset_notification(1, 4)])
    await mock_gateway.check_single_write(
        DynetPacket.select_area_preset_packet(1, 4, 0)
    )
    assert not room_device.is_on
    assert not on_device.is_on
    assert off_device.is_on
    room_device.init_level(2)
    assert room_device.is_on
    assert on_device.is_on
    assert not off_device.is_on
    room_device.init_level(0)
    assert not room_device.is_on
    assert not on_device.is_on
    assert off_device.is_on


@pytest.mark.asyncio
async def test_trigger_switch(mock_gateway):
    """Test a switch that is a single trigger."""
    name = "NAME"
    [trigger_device] = mock_gateway.configure_dyn_dev(
        {
            dyn_const.CONF_ACTIVE: False,
            dyn_const.CONF_AREA: {
                "1": {
                    dyn_const.CONF_NAME: name,
                    dyn_const.CONF_TEMPLATE: dyn_const.CONF_TRIGGER,
                }
            },
        },
        1,
    )
    assert await mock_gateway.async_setup_dyn_dev()
    await mock_gateway.check_single_update(None)
    assert trigger_device.category == "switch"
    assert trigger_device.name == name
    assert trigger_device.unique_id == "dynalite_area_1_preset_1"
    assert trigger_device.available
    await trigger_device.async_turn_on()
    await mock_gateway.check_single_write(
        DynetPacket.select_area_preset_packet(1, 1, 0)
    )
    await mock_gateway.check_single_update(trigger_device)
    await mock_gateway.check_notifications([preset_notification(1, 1)])
    assert trigger_device.is_on
    packet_to_send = DynetPacket.report_area_preset_packet(1, 4)
    await mock_gateway.receive(packet_to_send)
    await mock_gateway.check_single_update(trigger_device)
    await mock_gateway.check_notifications(
        [packet_notification(packet_to_send.raw_msg), preset_notification(1, 4)]
    )
    assert not trigger_device.is_on



"""Tests for DynaliteDevices."""

import pytest

import dynalite_devices_lib.const as dyn_const
from dynalite_devices_lib.dynet import DynetPacket
from dynalite_devices_lib.opcodes import OpcodeType

from .common import packet_notification, preset_notification


def preset_select_func(area, preset):
    """Create preset selection packet."""
    return DynetPacket.select_area_preset_packet(area, preset, 0)


def linear_func(area, preset):
    """Create preset linear fade packet."""
    return DynetPacket(
        area=area, command=OpcodeType.LINEAR_PRESET.value, data=[preset - 1, 0, 0]
    )


def report_func(area, preset):
    """Create preset report packet."""
    return DynetPacket.report_area_preset_packet(area, preset)


def set_channel_func(area, channel):
    """Create channel set level packet."""
    return DynetPacket.set_channel_level_packet(area, channel, 1, 0)


def report_channel_func(area, channel):
    """Create channel report level packet."""
    return DynetPacket.report_channel_level_packet(area, channel, 1, 1)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "conf, packet_func",
    [
        (dyn_const.CONF_PRESET, preset_select_func),
        (dyn_const.CONF_PRESET, linear_func),
        (dyn_const.CONF_PRESET, report_func),
        (dyn_const.CONF_CHANNEL, set_channel_func),
        (dyn_const.CONF_CHANNEL, report_channel_func),
    ],
)
async def test_selections(mock_gateway, conf, packet_func):
    """Run preset / channel selection tests with various commands."""
    devices = mock_gateway.configure_dyn_dev(
        {
            dyn_const.CONF_ACTIVE: False,
            dyn_const.CONF_AREA: {"2": {conf: {i: {} for i in range(1, 9)}}},
            dyn_const.CONF_PRESET: {},
        },
        8,
    )
    for device in devices:
        assert not device.is_on
    assert await mock_gateway.async_setup_dyn_dev()
    await mock_gateway.check_single_update(None)
    for i in range(1, 9):
        packet = packet_func(2, i)
        await mock_gateway.receive(packet)
        exp_notifications = [packet_notification(packet.raw_msg)]
        if conf == dyn_const.CONF_CHANNEL:
            await mock_gateway.check_single_update(devices[i - 1])
            assert devices[i - 1].is_on
        else:  # CONF_PRESET
            await mock_gateway.check_updates(devices)
            for j in range(1, 9):
                assert devices[j - 1].is_on == (i == j)
            exp_notifications.append(preset_notification(2, i))
        await mock_gateway.check_notifications(exp_notifications)


@pytest.mark.asyncio
async def test_inbound_request_channel_level(mock_gateway):
    """Test when the network requests a channel level. Nothing to do, just be sure nothing bad happens..."""
    [device] = mock_gateway.configure_dyn_dev(
        {
            dyn_const.CONF_ACTIVE: False,
            dyn_const.CONF_AREA: {"3": {dyn_const.CONF_CHANNEL: {"5": {}}}},
            dyn_const.CONF_PRESET: {},
        },
        1,
    )
    assert not device.is_on
    assert await mock_gateway.async_setup_dyn_dev()
    await mock_gateway.check_single_update(None)
    packet = DynetPacket.request_channel_level_packet(3, 5)
    await mock_gateway.receive(packet)
    await mock_gateway.check_notifications([packet_notification(packet.raw_msg)])
    assert not device.is_on



"""Tests for various packets in DynetPacket."""

import pytest

from dynalite_devices_lib.dynet import DynetPacket, PacketError


def test_packet_lengths():
    """Test what happens when creating a packet of size not equal to 8."""
    packet = DynetPacket(area=3, command=6, data=[1, 2, 3])
    with pytest.raises(PacketError):
        DynetPacket(packet.msg[0:7])
    with pytest.raises(PacketError):
        DynetPacket(packet.msg + bytearray([3]))
    DynetPacket(packet.msg[0:8])


def test_channel_fade_limit():
    """Test that set_channel_level has a limit at 0xFF (5.1 seconds)."""
    packet1 = DynetPacket.set_channel_level_packet(1, 1, 1, 100)
    packet2 = DynetPacket.set_channel_level_packet(1, 1, 1, 5.101)
    packet3 = DynetPacket.set_channel_level_packet(1, 1, 1, 5.05)
    assert packet1.msg == packet2.msg
    assert packet1.msg != packet3.msg



"""Tests for Dynalite lights."""
import pytest

import dynalite_devices_lib.const as dyn_const
from dynalite_devices_lib.dynet import DynetPacket

from .common import packet_notification


@pytest.mark.asyncio
async def test_light(mock_gateway):
    """Test a Dynalite channel that is a light."""
    name = "NAME"
    channel_name = "CHANNEL"
    [device] = mock_gateway.configure_dyn_dev(
        {
            dyn_const.CONF_ACTIVE: False,
            dyn_const.CONF_AREA: {
                "1": {
                    dyn_const.CONF_NAME: name,
                    dyn_const.CONF_NO_DEFAULT: True,
                    dyn_const.CONF_CHANNEL: {
                        "1": {
                            dyn_const.CONF_NAME: channel_name,
                            dyn_const.CONF_FADE: 0.5,
                        }
                    },
                }
            },
        }
    )
    assert await mock_gateway.async_setup_dyn_dev()
    await mock_gateway.check_single_update(None)
    assert device.category == "light"
    assert device.name == f"{name} {channel_name}"
    assert device.unique_id == "dynalite_area_1_channel_1"
    assert device.available
    assert device.area_name == name
    assert device.get_master_area == name
    await device.async_turn_on()
    await mock_gateway.check_single_write(
        DynetPacket.set_channel_level_packet(1, 1, 1.0, 0.5)
    )
    await mock_gateway.check_single_update(device)
    assert device.brightness == 255
    await device.async_turn_on(brightness=51)
    await mock_gateway.check_single_write(
        DynetPacket.set_channel_level_packet(1, 1, 0.2, 0.5)
    )
    await mock_gateway.check_single_update(device)
    assert device.brightness == 51
    await device.async_turn_off()
    await mock_gateway.check_single_write(
        DynetPacket.set_channel_level_packet(1, 1, 0, 0.5)
    )
    await mock_gateway.check_single_update(device)
    assert device.brightness == 0
    # Now send commands
    packet_to_send = DynetPacket.set_channel_level_packet(1, 1, 1.0, 0.5)
    await mock_gateway.receive(packet_to_send)
    await mock_gateway.check_single_update(device)
    await mock_gateway.check_notifications(
        [packet_notification(packet_to_send.raw_msg)]
    )
    assert device.brightness == 255
    assert device.is_on

    packet_to_send = DynetPacket.set_channel_level_packet(1, 1, 0.2, 0.5)
    await mock_gateway.receive(packet_to_send)
    await mock_gateway.check_single_update(device)
    await mock_gateway.check_notifications(
        [packet_notification(packet_to_send.raw_msg)]
    )
    assert device.brightness == 51
    assert device.is_on

    packet_to_send = DynetPacket.report_channel_level_packet(1, 1, 0, 0)
    await mock_gateway.receive(packet_to_send)
    await mock_gateway.check_single_update(device)
    await mock_gateway.check_notifications(
        [packet_notification(packet_to_send.raw_msg)]
    )
    assert device.brightness == 0
    assert not device.is_on


@pytest.mark.asyncio
async def test_light_to_preset(mock_gateway):
    """Test a Dynalite channel that is a light."""
    name = "NAME"
    channel_name = "CHANNEL"
    [device, preset_1, preset_2, preset_3] = mock_gateway.configure_dyn_dev(
        {
            dyn_const.CONF_ACTIVE: False,
            dyn_const.CONF_AREA: {
                "1": {
                    dyn_const.CONF_NAME: name,
                    dyn_const.CONF_CHANNEL: {
                        "1": {
                            dyn_const.CONF_NAME: channel_name,
                            dyn_const.CONF_FADE: 0.5,
                        }
                    },
                    dyn_const.CONF_PRESET: {"2": {dyn_const.CONF_LEVEL: 0.2}},
                }
            },
        },
        4,
    )
    assert await mock_gateway.async_setup_dyn_dev()
    await mock_gateway.check_single_update(None)
    assert device.category == "light"
    assert device.name == f"{name} {channel_name}"
    assert device.unique_id == "dynalite_area_1_channel_1"
    assert device.available
    assert device.area_name == name
    assert device.get_master_area == name
    # Now send commands
    packet_to_send = DynetPacket.fade_area_channel_preset_packet(1, 1, 2, 0.0)
    await mock_gateway.receive(packet_to_send)
    await mock_gateway.check_single_update(device)
    await mock_gateway.check_notifications(
        [packet_notification(packet_to_send.raw_msg)]
    )
    assert device.brightness == 51
    assert device.is_on
    # check default preset on
    packet_to_send = DynetPacket.fade_area_channel_preset_packet(1, 1, 1, 0.0)
    await mock_gateway.receive(packet_to_send)
    await mock_gateway.check_single_update(device)
    await mock_gateway.check_notifications(
        [packet_notification(packet_to_send.raw_msg)]
    )
    assert device.brightness == 255
    assert device.is_on
    # check default preset off
    packet_to_send = DynetPacket.fade_area_channel_preset_packet(1, 1, 4, 0.0)
    await mock_gateway.receive(packet_to_send)
    await mock_gateway.check_single_update(device)
    await mock_gateway.check_notifications(
        [packet_notification(packet_to_send.raw_msg)]
    )
    assert device.brightness == 0
    assert not device.is_on


@pytest.mark.asyncio
async def test_light_init(mock_gateway):
    """Test a Dynalite light getting initial value."""
    [device, _, _] = mock_gateway.configure_dyn_dev(
        {
            dyn_const.CONF_ACTIVE: False,
            dyn_const.CONF_AREA: {
                "1": {
                    dyn_const.CONF_CHANNEL: {"1": {}},
                }
            },
        },
        3,
    )
    assert await mock_gateway.async_setup_dyn_dev()
    await mock_gateway.check_single_update(None)
    assert device.available
    # Now send commands
    device.init_level(135)
    assert device.brightness == 135
    assert device.is_on
    device.init_level(0)
    assert device.brightness == 0
    assert not device.is_on
    with pytest.raises(ValueError):
        device.init_level(-1)
    with pytest.raises(ValueError):
        device.init_level(256)



"""Configure the areas, presets, and channels."""

from typing import Any, Dict, Union

from .const import (
    ACTIVE_INIT,
    ACTIVE_OFF,
    ACTIVE_ON,
    CONF_ACTIVE,
    CONF_AREA,
    CONF_AREA_OVERRIDE,
    CONF_AUTO_DISCOVER,
    CONF_CHANNEL,
    CONF_CHANNEL_COVER,
    CONF_CHANNEL_TYPE,
    CONF_CLOSE_PRESET,
    CONF_DEFAULT,
    CONF_DEVICE_CLASS,
    CONF_DURATION,
    CONF_FADE,
    CONF_HIDDEN_ENTITY,
    CONF_HOST,
    CONF_LEVEL,
    CONF_NAME,
    CONF_NO_DEFAULT,
    CONF_OPEN_PRESET,
    CONF_POLL_TIMER,
    CONF_PORT,
    CONF_PRESET,
    CONF_QUERY_CHANNEL,
    CONF_ROOM,
    CONF_ROOM_OFF,
    CONF_ROOM_ON,
    CONF_STOP_PRESET,
    CONF_TEMPLATE,
    CONF_TILT_TIME,
    CONF_TIME_COVER,
    CONF_TRIGGER,
    DEFAULT_CHANNEL_TYPE,
    DEFAULT_NAME,
    DEFAULT_PORT,
    DEFAULT_PRESETS,
    DEFAULT_QUERY_CHANNEL,
    DEFAULT_TEMPLATES,
)

PRESET_CONFS = {
    CONF_ROOM: [CONF_ROOM_ON, CONF_ROOM_OFF],
    CONF_TRIGGER: [CONF_TRIGGER],
    CONF_TIME_COVER: [CONF_OPEN_PRESET, CONF_CLOSE_PRESET, CONF_STOP_PRESET],
}

TIME_COVER_VALUE_CONFS = [CONF_DEVICE_CLASS, CONF_DURATION, CONF_TILT_TIME]


class DynaliteConfig:
    """Configure the Dynalite bridge."""

    def __init__(self, config: Dict[str, Any]) -> None:
        """Configure the Dynalite bridge."""
        # insert the global values
        self.host = config.get(CONF_HOST, "localhost")  # Default value for testing
        self.port = config.get(CONF_PORT, DEFAULT_PORT)
        self.name = config.get(CONF_NAME, f"{DEFAULT_NAME}-{self.host}")
        self.auto_discover = config.get(CONF_AUTO_DISCOVER, False)
        temp_active = config.get(CONF_ACTIVE, ACTIVE_INIT)
        if temp_active is True:
            self.active = ACTIVE_ON
        elif temp_active is False:
            self.active = ACTIVE_OFF
        else:
            self.active = temp_active
        self.poll_timer = config.get(CONF_POLL_TIMER, 1.0)
        self.default_fade = config.get(CONF_DEFAULT, {}).get(CONF_FADE, 0)
        self.default_query_channel = int(
            config.get(CONF_DEFAULT, {}).get(CONF_QUERY_CHANNEL, DEFAULT_QUERY_CHANNEL)
        )
        # create the templates
        config_templates = config.get(CONF_TEMPLATE, {})
        templates: Dict[str, Dict[str, Union[str, int]]] = {}
        # for template in DEFAULT_TEMPLATES:
        for template in DEFAULT_TEMPLATES:
            templates[template] = {}
            cur_template = config_templates.get(template, {})
            for conf in DEFAULT_TEMPLATES[template]:
                templates[template][conf] = cur_template.get(
                    conf, DEFAULT_TEMPLATES[template][conf]
                )
        # create default presets
        config_presets = config.get(CONF_PRESET, DEFAULT_PRESETS)
        self.default_presets = {
            int(preset): self.configure_preset(
                preset, config_presets[preset], self.default_fade
            )
            for preset in config_presets
        }
        # create the areas with their channels and presets
        self.area = {}
        for area_val in config.get(CONF_AREA, {}):  # may be a string '123'
            area = int(area_val)
            area_config = config[CONF_AREA].get(area_val)
            self.area[area] = self.configure_area(
                area,
                area_config,
                self.default_fade,
                self.default_query_channel,
                templates,
                self.default_presets,
            )

    @staticmethod
    def configure_preset(
        preset: int,
        preset_config: Dict[str, Union[float, str]],
        default_fade: float,
        hidden: bool = False,
    ) -> Dict[str, Union[str, float, bool]]:
        """Return the configuration of a preset."""
        result = {
            CONF_NAME: preset_config.get(CONF_NAME, f"Preset {preset}"),
            CONF_FADE: preset_config.get(CONF_FADE, default_fade),
        }
        if CONF_LEVEL in preset_config:
            result[CONF_LEVEL] = preset_config[CONF_LEVEL]
        if hidden:
            result[CONF_HIDDEN_ENTITY] = True
        return result

    @staticmethod
    def configure_channel(
        channel: int,
        channel_config: Dict[str, Union[float, str]],
        default_fade: float,
        hidden: bool = False,
    ) -> Dict[str, Union[str, float, bool]]:
        """Return the configuration of a channel."""
        result = {
            CONF_NAME: channel_config.get(CONF_NAME, f"Channel {channel}"),
            CONF_FADE: channel_config.get(CONF_FADE, default_fade),
            CONF_CHANNEL_TYPE: channel_config.get(
                CONF_CHANNEL_TYPE, DEFAULT_CHANNEL_TYPE
            ),
        }
        if hidden:
            result[CONF_HIDDEN_ENTITY] = True
        return result

    @staticmethod
    def configure_area(
        area: int,
        area_config: Dict[str, Any],
        default_fade: float,
        default_query_channel: int,
        templates: Dict[str, Dict[str, Union[str, int]]],
        default_presets: Dict[int, Any],
    ) -> Dict[str, Any]:
        """Return the configuration of an area."""
        result = {
            CONF_NAME: area_config.get(CONF_NAME, f"Area {area}"),
            CONF_FADE: area_config.get(CONF_FADE, default_fade),
            CONF_QUERY_CHANNEL: int(
                area_config.get(CONF_QUERY_CHANNEL, default_query_channel)
            ),
        }
        for conf in [CONF_TEMPLATE, CONF_AREA_OVERRIDE]:
            if conf in area_config:
                result[conf] = area_config[conf]
        # User defined presets and channels first, then template presets, then defaults
        area_presets = {
            int(preset): DynaliteConfig.configure_preset(
                preset, area_config[CONF_PRESET][preset], result[CONF_FADE]
            )
            for preset in area_config.get(CONF_PRESET, {})
        }
        area_channels = {
            int(channel): DynaliteConfig.configure_channel(
                channel, area_config[CONF_CHANNEL][channel], result[CONF_FADE]
            )
            for channel in area_config.get(CONF_CHANNEL, {})
        }
        # add the entities implicitly defined by templates
        template = area_config.get(CONF_TEMPLATE)
        if template:
            # ensure presets are there
            for conf in PRESET_CONFS[template]:
                preset = int(area_config.get(conf, templates[template][conf]))
                result[conf] = preset
                if preset not in area_presets:
                    if template == CONF_TRIGGER:
                        area_presets[preset] = DynaliteConfig.configure_preset(
                            preset,
                            {CONF_NAME: result[CONF_NAME]},
                            result[CONF_FADE],
                            False,
                        )
                    else:
                        area_presets[preset] = DynaliteConfig.configure_preset(
                            preset, {}, result[CONF_FADE], True
                        )
        if template == CONF_TIME_COVER:  # time cover also has non-preset conf
            for conf in TIME_COVER_VALUE_CONFS:
                result[conf] = area_config.get(conf, templates[template][conf])
            channel_cover = int(
                area_config.get(
                    CONF_CHANNEL_COVER, templates[template][CONF_CHANNEL_COVER]
                )
            )
            result[CONF_CHANNEL_COVER] = channel_cover
            if 0 < channel_cover < 255 and channel_cover not in area_channels:
                area_channels[channel_cover] = DynaliteConfig.configure_channel(
                    channel_cover, {}, result[CONF_FADE], True
                )
        # Default presets
        if not area_config.get(CONF_NO_DEFAULT, False) and not template:
            for preset in default_presets:
                if preset not in area_presets:
                    area_presets[preset] = default_presets[preset]
        result[CONF_PRESET] = area_presets
        result[CONF_CHANNEL] = area_channels
        return result



"""Tests for DynaliteDevices."""

import asyncio
from unittest.mock import patch

import pytest

import dynalite_devices_lib.const as dyn_const
from dynalite_devices_lib.dynet import DynetPacket
from dynalite_devices_lib.opcodes import SyncType

from .common import packet_notification


@pytest.mark.asyncio
async def test_dynalite_disconnection(mock_gateway):
    """Test a network disconnection."""
    devices = mock_gateway.configure_dyn_dev(
        {
            dyn_const.CONF_ACTIVE: False,
            dyn_const.CONF_AREA: {
                "1": {dyn_const.CONF_CHANNEL: {"1": {}, "2": {}}},
                "2": {dyn_const.CONF_TEMPLATE: dyn_const.CONF_ROOM},
                "3": {dyn_const.CONF_TEMPLATE: dyn_const.CONF_TIME_COVER},
            },
            dyn_const.CONF_PRESET: {"6": {}},
        },
        5,
    )
    assert await mock_gateway.async_setup_dyn_dev()
    await mock_gateway.check_single_update(None)
    for device in devices:
        assert device.available
    # Disconnect
    with patch("dynalite_devices_lib.dynalite.CONNECTION_RETRY_DELAY", 0.1):
        await mock_gateway.shutdown()
        await asyncio.sleep(0.05)
        await mock_gateway.check_single_update(None)
        for device in devices:
            assert not device.available
        await asyncio.sleep(0.2)
        await mock_gateway.async_setup_server()
        # Wait for reconnect
        await asyncio.sleep(0.2)
        await mock_gateway.check_single_update(None)
        for device in devices:
            assert device.available


@pytest.mark.asyncio
async def test_dynalite_connection_reset(mock_gateway):
    """Test a connection reset."""
    devices = mock_gateway.configure_dyn_dev(
        {
            dyn_const.CONF_ACTIVE: True,
            dyn_const.CONF_AREA: {
                "1": {dyn_const.CONF_CHANNEL: {"1": {}, "2": {}}},
                "2": {dyn_const.CONF_TEMPLATE: dyn_const.CONF_ROOM},
                "3": {dyn_const.CONF_TEMPLATE: dyn_const.CONF_TIME_COVER},
            },
            dyn_const.CONF_PRESET: {"6": {}},
        },
        5,
    )
    assert await mock_gateway.async_setup_dyn_dev()
    await mock_gateway.check_single_update(None)
    for device in devices:
        assert device.available
    # Disconnect
    with patch("dynalite_devices_lib.dynalite.CONNECTION_RETRY_DELAY", 0.1):
        # abort instead of close causes the connection to be reset
        writer = mock_gateway.writer
        writer.transport.abort()
        await mock_gateway.shutdown()
        await asyncio.sleep(0.05)
        await mock_gateway.check_single_update(None)
        for device in devices:
            assert not device.available
        await mock_gateway.async_setup_server()
        await asyncio.sleep(0.1)
        await mock_gateway.check_single_update(None)
        for device in devices:
            assert device.available


@pytest.mark.asyncio
async def test_dynalite_no_server(mock_gateway):
    """Test when no server is configured."""
    mock_gateway.configure_dyn_dev({dyn_const.CONF_PORT: 12333}, 0)
    assert not await mock_gateway.async_setup_dyn_dev()


@pytest.mark.asyncio
async def test_dynalite_shutdown_with_server_down(mock_gateway):
    """Test when shutting down while the server is down. Different flow in async_reset."""
    [device] = mock_gateway.configure_dyn_dev(
        {
            dyn_const.CONF_ACTIVE: False,
            dyn_const.CONF_AREA: {"1": {dyn_const.CONF_CHANNEL: {"1": {}}}},
            dyn_const.CONF_PRESET: {},
        },
    )
    assert await mock_gateway.async_setup_dyn_dev()
    await mock_gateway.check_single_update(None)
    assert device.available
    # Disconnect
    await mock_gateway.shutdown()
    await asyncio.sleep(0.1)
    await mock_gateway.check_single_update(None)
    assert not device.available


@pytest.mark.asyncio
async def test_dynalite_split_message(mock_gateway):
    """Test when a received message is split into two packets."""
    [device] = mock_gateway.configure_dyn_dev(
        {
            dyn_const.CONF_ACTIVE: False,
            dyn_const.CONF_AREA: {"1": {dyn_const.CONF_CHANNEL: {"1": {}}}},
            dyn_const.CONF_PRESET: {},
        },
    )
    assert await mock_gateway.async_setup_dyn_dev()
    await mock_gateway.check_single_update(None)
    assert not device.is_on
    packet = DynetPacket.set_channel_level_packet(1, 1, 1.0, 0.5)
    msg1 = packet.msg[0:5]
    msg2 = packet.msg[5:8]
    await mock_gateway.receive_message(msg1)
    await mock_gateway.receive_message(msg2)
    await mock_gateway.check_single_update(device)
    await mock_gateway.check_notifications([packet_notification(packet.raw_msg)])
    assert device.is_on


@pytest.mark.asyncio
async def test_dynalite_shift_message(mock_gateway):
    """Test when a received message arrives out of the normal 8bit pattern."""
    [device] = mock_gateway.configure_dyn_dev(
        {
            dyn_const.CONF_ACTIVE: False,
            dyn_const.CONF_AREA: {"1": {dyn_const.CONF_CHANNEL: {"1": {}}}},
            dyn_const.CONF_PRESET: {},
        },
    )
    assert await mock_gateway.async_setup_dyn_dev()
    await mock_gateway.check_single_update(None)
    assert not device.is_on
    packet = DynetPacket.set_channel_level_packet(1, 1, 1.0, 0.5)
    message = bytearray([3, 7, 12]) + packet.msg
    await mock_gateway.receive_message(message)
    await mock_gateway.check_single_update(device)
    await mock_gateway.check_notifications([packet_notification(packet.raw_msg)])
    assert device.is_on


@pytest.mark.asyncio
async def test_dynalite_debug_message(mock_gateway):
    """Test when a DEBUG message arrives."""
    [device] = mock_gateway.configure_dyn_dev(
        {
            dyn_const.CONF_ACTIVE: False,
            dyn_const.CONF_AREA: {"1": {dyn_const.CONF_CHANNEL: {"1": {}}}},
            dyn_const.CONF_PRESET: {},
        },
    )
    assert await mock_gateway.async_setup_dyn_dev()
    await mock_gateway.check_single_update(None)
    assert not device.is_on
    # putting a debug message + a device-on message half-way.
    # will verify that it will first read the debug message and then ignore other
    packet = DynetPacket.set_channel_level_packet(1, 1, 1.0, 0.5)
    int_message = [SyncType.DEBUG_MSG.value, 64, 65] + packet.raw_msg
    message = bytearray(int_message)
    await mock_gateway.receive_message(message)
    await mock_gateway.check_notifications([packet_notification(int_message[:8])])
    assert not device.is_on
    await mock_gateway.receive(packet)
    await mock_gateway.check_single_update(device)
    await mock_gateway.check_notifications([packet_notification(packet.raw_msg)])
    assert device.is_on


@pytest.mark.asyncio
async def test_dynalite_device_message(mock_gateway):
    """Test when a DEVICE message arrives."""
    [device] = mock_gateway.configure_dyn_dev(
        {
            dyn_const.CONF_ACTIVE: False,
            dyn_const.CONF_AREA: {"1": {dyn_const.CONF_CHANNEL: {"1": {}}}},
            dyn_const.CONF_PRESET: {},
        },
    )
    assert await mock_gateway.async_setup_dyn_dev()
    await mock_gateway.check_single_update(None)
    assert not device.is_on
    # putting a device message + a device-on message half-way.
    # will verify that it will first read the device message and then ignore other
    packet = DynetPacket.set_channel_level_packet(1, 1, 1.0, 0.5)
    int_message = [SyncType.DEVICE.value, 65, 66] + packet.raw_msg
    message = bytearray(int_message)
    await mock_gateway.receive_message(message)
    await mock_gateway.check_notifications([packet_notification(int_message[:8])])
    assert not device.is_on
    await mock_gateway.receive(packet)
    await mock_gateway.check_single_update(device)
    await mock_gateway.check_notifications([packet_notification(packet.raw_msg)])
    assert device.is_on


@pytest.mark.asyncio
async def test_dynalite_error_message(mock_gateway):
    """Test when a message arrives with the wrong checksum."""
    [device] = mock_gateway.configure_dyn_dev(
        {
            dyn_const.CONF_ACTIVE: False,
            dyn_const.CONF_AREA: {"1": {dyn_const.CONF_CHANNEL: {"1": {}}}},
            dyn_const.CONF_PRESET: {},
        },
    )
    assert await mock_gateway.async_setup_dyn_dev()
    await mock_gateway.check_single_update(None)
    assert not device.is_on
    packet = DynetPacket.set_channel_level_packet(1, 1, 1.0, 0.5)
    message = packet.msg
    message[7] += 1
    await mock_gateway.receive_message(message)
    assert not device.is_on
    packet = DynetPacket.set_channel_level_packet(1, 1, 1.0, 0.5)
    await mock_gateway.receive(packet)
    await mock_gateway.check_single_update(device)
    await mock_gateway.check_notifications([packet_notification(packet.raw_msg)])
    assert device.is_on


@pytest.mark.asyncio
async def test_dynalite_unhandled_message(mock_gateway):
    """Test when a message arrives that we know but don't handle."""
    [device] = mock_gateway.configure_dyn_dev(
        {
            dyn_const.CONF_ACTIVE: False,
            dyn_const.CONF_AREA: {"1": {dyn_const.CONF_CHANNEL: {"1": {}}}},
            dyn_const.CONF_PRESET: {},
        },
    )
    assert await mock_gateway.async_setup_dyn_dev()
    await mock_gateway.check_single_update(None)
    assert not device.is_on
    packet = DynetPacket(area=1, command=45, data=[0, 0, 0])
    await mock_gateway.receive(packet)
    await mock_gateway.check_notifications([packet_notification(packet.raw_msg)])
    assert not device.is_on
    packet = DynetPacket.set_channel_level_packet(1, 1, 1.0, 0.5)
    await mock_gateway.receive(packet)
    await mock_gateway.check_single_update(device)
    await mock_gateway.check_notifications([packet_notification(packet.raw_msg)])
    assert device.is_on


@pytest.mark.asyncio
async def test_dynalite_unknown_message(mock_gateway):
    """Test when a message arrives that we don't know."""
    [device] = mock_gateway.configure_dyn_dev(
        {
            dyn_const.CONF_ACTIVE: False,
            dyn_const.CONF_AREA: {"1": {dyn_const.CONF_CHANNEL: {"1": {}}}},
            dyn_const.CONF_PRESET: {},
        },
    )
    assert await mock_gateway.async_setup_dyn_dev()
    await mock_gateway.check_single_update(None)
    assert not device.is_on
    packet = DynetPacket(area=1, command=200, data=[0, 0, 0])
    await mock_gateway.receive(packet)
    await mock_gateway.check_notifications([packet_notification(packet.raw_msg)])
    assert not device.is_on
    packet = DynetPacket.set_channel_level_packet(1, 1, 1.0, 0.5)
    await mock_gateway.receive(packet)
    await mock_gateway.check_single_update(device)
    await mock_gateway.check_notifications([packet_notification(packet.raw_msg)])
    assert device.is_on


@pytest.mark.asyncio
async def test_dynalite_two_messages(mock_gateway):
    """Test when two messages arrive together."""
    devices = mock_gateway.configure_dyn_dev(
        {
            dyn_const.CONF_ACTIVE: False,
            dyn_const.CONF_AREA: {"1": {dyn_const.CONF_CHANNEL: {"1": {}, "2": {}}}},
            dyn_const.CONF_PRESET: {},
        },
        2,
    )
    assert await mock_gateway.async_setup_dyn_dev()
    await mock_gateway.check_single_update(None)
    for device in devices:
        assert not device.is_on
    packet1 = DynetPacket.set_channel_level_packet(1, 1, 1.0, 0.5)
    packet2 = DynetPacket.set_channel_level_packet(1, 2, 1.0, 0.5)
    await mock_gateway.receive_message(packet1.msg + packet2.msg)
    await mock_gateway.check_updates(devices)
    await mock_gateway.check_notifications(
        [packet_notification(packet1.raw_msg), packet_notification(packet2.raw_msg)]
    )
    for device in devices:
        assert device.is_on


@pytest.mark.asyncio
async def test_dynalite_write_message_throttle(mock_gateway_with_delay):
    """Test that when we send many messages, it throttles the messages."""
    mock_gateway_with_delay.configure_dyn_dev(
        {
            dyn_const.CONF_ACTIVE: True,
            dyn_const.CONF_AREA: {
                i: {dyn_const.CONF_CHANNEL: {1: {}}} for i in range(1, 26)
            },
            dyn_const.CONF_PRESET: {},
        },
        25,
    )
    assert await mock_gateway_with_delay.async_setup_dyn_dev()
    await mock_gateway_with_delay.check_single_update(None)
    await asyncio.sleep(1)  # should be roughly 5 messages
    assert 3 * 8 <= len(mock_gateway_with_delay.in_buffer) <= 7 * 8



"""
Manage a Dynalite connection.

@ Author      : Troy Kelly
@ Date        : 3 Dec 2018
@ Description : Philips Dynalite Library - Unofficial interface for Philips Dynalite over RS485

@ Notes:        Requires a RS485 to IP gateway (Do not use the Dynalite one - use something cheaper)
"""

import asyncio
import time
from typing import Awaitable, Callable, List, Optional

from .const import (
    CONF_ACTION,
    CONF_ACTION_CMD,
    CONF_AREA,
    CONF_CHANNEL,
    CONF_PRESET,
    CONF_TRGT_LEVEL,
    CONNECTION_RETRY_DELAY,
    EVENT_CHANNEL,
    EVENT_CONNECTED,
    EVENT_DISCONNECTED,
    EVENT_PACKET,
    EVENT_PRESET,
    LOGGER,
    MESSAGE_DELAY,
)
from .dynet import DynetPacket, PacketError
from .event import DynetEvent
from .inbound import DynetInbound
from .opcodes import SyncType


class Dynalite:
    """Class to represent the interaction with Dynalite."""

    def __init__(self, broadcast_func: Callable[[DynetEvent], None]) -> None:
        """Initialize the class."""
        self._loop: Optional[asyncio.AbstractEventLoop] = None
        self._broadcast_func = broadcast_func
        self._in_buffer: List[int] = []
        self._out_buffer: List[DynetPacket] = []
        self._last_sent = 0.0
        self._message_delay = MESSAGE_DELAY  # public for testing
        self._reader: Optional[asyncio.StreamReader] = None
        self._writer: Optional[asyncio.StreamWriter] = None
        self._resetting = False
        self._reader_future: Optional[Awaitable[None]] = None

    async def connect_internal(self, host: str, port: int) -> bool:
        """Create the actual connection to Dynet."""
        try:
            self._reader, self._writer = await asyncio.open_connection(host, port)
            return True
        except (ValueError, OSError, asyncio.TimeoutError) as err:
            LOGGER.warning("Could not connect to Dynet (%s)", err)
            return False

    async def connect(self, host: str, port: int) -> bool:
        """Connect to Dynet."""
        LOGGER.debug("Connecting to Dynet on %s:%s", host, port)
        self._loop = asyncio.get_running_loop()
        self._resetting = False
        result = await self.connect_internal(host, port)
        if result and not self._resetting:
            self._reader_future = self._loop.create_task(self.reader_loop(host, port))
            self.broadcast(DynetEvent(event_type=EVENT_CONNECTED))
        return result

    async def reader_loop(self, host: str, port: int) -> None:
        """Loop to read from the stream reader and reconnect if necessary."""
        while True:
            self.write()  # write if there is something in the buffers
            try:
                assert self._reader
                data = await self._reader.read(100)
                if len(data) > 0:
                    self.receive(data)
                    continue
            except ConnectionResetError:
                pass
            # we got disconnected or EOF
            if self._resetting:
                self._reader = None
                return  # stop loop
            self._reader = None
            self._writer = None
            self.broadcast(DynetEvent(event_type=EVENT_DISCONNECTED))
            await asyncio.sleep(CONNECTION_RETRY_DELAY)  # Don't overload the network
            while not await self.connect_internal(host, port):
                if self._resetting:
                    self._reader = None
                    return  # stop loop
                # Don't overload the network
                await asyncio.sleep(CONNECTION_RETRY_DELAY)
            self.broadcast(DynetEvent(event_type=EVENT_CONNECTED))

    def broadcast(self, event: DynetEvent) -> None:
        """Broadcast an event to all listeners - queue."""
        assert self._loop
        self._loop.call_soon(self._broadcast_func, event)

    def set_channel_level(
        self, area: int, channel: int, level: float, fade: float
    ) -> None:
        """Set the level of a channel."""
        packet = DynetPacket.set_channel_level_packet(area, channel, level, fade)
        self.write(packet)
        broadcast_data = {
            CONF_AREA: area,
            CONF_CHANNEL: channel,
            CONF_TRGT_LEVEL: int(255 - 254.0 * level),
            CONF_ACTION: CONF_ACTION_CMD,
        }
        self.broadcast(DynetEvent(event_type=EVENT_CHANNEL, data=broadcast_data))

    def select_preset(self, area: int, preset: int, fade: float) -> None:
        """Select a preset in an area."""
        packet = DynetPacket.select_area_preset_packet(area, preset, fade)
        self.write(packet)
        broadcast_data = {
            CONF_AREA: area,
            CONF_PRESET: preset,
        }
        self.broadcast(DynetEvent(event_type=EVENT_PRESET, data=broadcast_data))

    def request_channel_level(self, area: int, channel: int) -> None:
        """Request a level for a specific channel."""
        packet = DynetPacket.request_channel_level_packet(area, channel)
        self.write(packet)

    def request_area_preset(self, area: int, query_channel: int) -> None:
        """Request current preset of an area."""
        packet = DynetPacket.request_area_preset_packet(area, query_channel)
        self.write(packet)

    def next_packet(self) -> Optional[DynetPacket]:
        """Get a valid packet from in_buffer."""
        packet = None
        while len(self._in_buffer) >= 8 and packet is None:
            first_byte = self._in_buffer[0]
            if SyncType.has_value(first_byte):
                if first_byte == SyncType.DEBUG_MSG.value:
                    bytemsg = "".join(chr(c) for c in self._in_buffer[1:7])
                    LOGGER.debug("Dynet DEBUG message %s", bytemsg)
                    self.broadcast(
                        DynetEvent(
                            event_type=EVENT_PACKET,
                            data={EVENT_PACKET: self._in_buffer[:8]},
                        )
                    )
                    self._in_buffer = self._in_buffer[8:]
                    continue
                if first_byte == SyncType.DEVICE.value:
                    LOGGER.debug(
                        "Not handling Dynet DEVICE message %s", self._in_buffer[:8]
                    )
                    self.broadcast(
                        DynetEvent(
                            event_type=EVENT_PACKET,
                            data={EVENT_PACKET: self._in_buffer[:8]},
                        )
                    )
                    self._in_buffer = self._in_buffer[8:]
                    continue
                assert first_byte == SyncType.LOGICAL.value
                try:
                    packet = DynetPacket(msg=self._in_buffer[:8])
                except PacketError as err:
                    LOGGER.warning(err)
                    packet = None
            if packet is None:
                hex_string = ":".join("{:02x}".format(c) for c in self._in_buffer[:8])
                LOGGER.debug(
                    "Unable to process packet %s - moving one byte forward", hex_string
                )
                del self._in_buffer[0]
                continue
            self.broadcast(
                DynetEvent(
                    event_type=EVENT_PACKET, data={EVENT_PACKET: self._in_buffer[:8]}
                )
            )
            self._in_buffer = self._in_buffer[8:]
        return packet

    @staticmethod
    def event_from_packet(packet: DynetPacket) -> Optional[DynetEvent]:
        """Create an event from a valid packet."""
        if packet.opcode_type is not None:
            inbound_handler = DynetInbound()
            if hasattr(inbound_handler, packet.opcode_type.lower()):
                event = getattr(inbound_handler, packet.opcode_type.lower())(packet)
                return event
            LOGGER.debug("Unhandled Dynet Inbound (%s): %s", packet.opcode_type, packet)
        else:
            LOGGER.debug("Unhandled Dynet Inbound: %s", packet)
        return None

    def receive(self, data: Optional[bytes] = None) -> None:
        """Handle data that was received."""
        if data is not None:
            for byte in data:
                self._in_buffer.append(int(byte))
        if len(self._in_buffer) < 8:
            LOGGER.debug(
                "Received %d bytes, not enough to process: %s",
                len(self._in_buffer),
                self._in_buffer,
            )
        packet = self.next_packet()
        if not packet:
            return
        LOGGER.debug("Have packet: %s", packet)
        event = self.event_from_packet(packet)
        if event:
            self.broadcast(event)
        # If there is still buffer to process - start again
        if len(self._in_buffer) >= 8:
            assert self._loop
            self._loop.call_soon(self.receive)

    def write(self, new_packet: Optional[DynetPacket] = None) -> None:
        """Write a packet or trigger write loop."""
        if new_packet is not None:
            self._out_buffer.append(new_packet)
        if self._writer is None:
            LOGGER.debug("write before transport is ready. queuing")
            return
        if self._message_delay > 0:  # in testing it is set to 0
            if self._last_sent == 0.0:
                self._last_sent = time.time()
            current_time = time.time()
            elapsed = current_time - self._last_sent
            delay = self._message_delay - elapsed
            if delay > 0:
                assert self._loop
                self._loop.call_later(delay, self.write)
                return
        if len(self._out_buffer) == 0:
            return
        packet = self._out_buffer[0]
        msg = packet.msg
        self._writer.write(msg)
        LOGGER.debug("Dynet Sent: %s", [int(byte) for byte in msg])
        self._last_sent = time.time()
        del self._out_buffer[0]
        if len(self._out_buffer) > 0:
            assert self._loop
            self._loop.call_later(self._message_delay, self.write)

    async def async_reset(self) -> None:
        """Close sockets and timers."""
        self._resetting = True
        # Wait for reader to also close
        if self._reader_future:
            await self._reader_future



"""Support for the Dynalite channels and presets as switches."""

from typing import TYPE_CHECKING

from .const import CONF_PRESET, CONF_ROOM, CONF_TEMPLATE
from .dynalitebase import (
    DynaliteBaseDevice,
    DynaliteChannelBaseDevice,
    DynaliteMultiDevice,
)

if TYPE_CHECKING:  # pragma: no cover
    from .dynalite_devices import DynaliteDevices


class DynaliteChannelSwitchDevice(DynaliteChannelBaseDevice):
    """Representation of a Dynalite Channel as a Home Assistant Switch."""

    def __init__(
        self, area: int, channel: int, bridge: "DynaliteDevices", hidden: bool
    ) -> None:
        """Initialize the switch."""
        self._level = 0.0
        super().__init__(area, channel, bridge, hidden)

    @property
    def category(self) -> str:
        """Return the category of the entity: light, switch, or cover."""
        return "switch"

    @property
    def is_on(self) -> bool:
        """Return true if switch is on."""
        return self._level > 0

    def update_level(self, actual_level: float, target_level: float) -> None:
        """Update the current level."""
        # pylint: disable=unused-argument
        self._level = actual_level

    async def async_turn_on(self, **kwargs) -> None:
        """Turn switch on."""
        # pylint: disable=unused-argument
        fade = self._bridge.get_channel_fade(self._area, self._channel)
        self._bridge.set_channel_level(self._area, self._channel, 1, fade)

    async def async_turn_off(self, **kwargs) -> None:
        """Turn switch off."""
        # pylint: disable=unused-argument
        fade = self._bridge.get_channel_fade(self._area, self._channel)
        self._bridge.set_channel_level(self._area, self._channel, 0, fade)

    def init_level(self, level):
        """Initialize to on/off."""
        if level > 0:
            self._level = 1.0
        elif level == 0:
            self._level = 0
        else:
            raise ValueError


class DynalitePresetSwitchDevice(DynaliteBaseDevice):
    """Representation of a Dynalite Preset as a Home Assistant Switch."""

    def __init__(
        self, area: int, preset: int, bridge: "DynaliteDevices", hidden: bool
    ) -> None:
        """Initialize the switch."""
        self._preset = preset
        self._level = 0
        super().__init__(area, bridge, hidden)

    @property
    def available(self) -> bool:
        """Return if device is available."""
        return self._bridge.available(CONF_PRESET, self._area, self._preset)

    @property
    def name(self) -> str:
        """Return the name of the device."""
        return self._bridge.get_preset_name(self._area, self._preset)

    @property
    def category(self) -> str:
        """Return the category of the entity: light, switch, or cover."""
        return "switch"

    @property
    def unique_id(self) -> str:
        """Return the ID of this cover."""
        return "dynalite_area_" + str(self._area) + "_preset_" + str(self._preset)

    def set_level(self, level: int) -> None:
        """Set the current level and potentially trigger listeners."""
        old_level = self._level
        self._level = level
        if old_level != self._level:
            self.update_listeners()

    @property
    def is_on(self) -> bool:
        """Return true if device is on."""
        return self._level > 0

    async def async_turn_on(self, **kwargs) -> None:
        """Turn switch on."""
        # pylint: disable=unused-argument
        fade = self._bridge.get_preset_fade(self._area, self._preset)
        self._bridge.select_preset(self._area, self._preset, fade)

    async def async_turn_off(self, **kwargs) -> None:
        """Turn switch off - doesn't do anything for presets."""
        # pylint: disable=unused-argument
        self.set_level(0)

    def init_level(self, level):
        """Initialize to on/off."""
        if level:
            self._level = 1.0
        else:
            self._level = 0


class DynaliteDualPresetSwitchDevice(DynaliteMultiDevice):
    """Representation of a Dynalite Preset as a Home Assistant Switch."""

    def __init__(self, area: int, bridge: "DynaliteDevices", hidden: bool) -> None:
        """Initialize the switch."""
        super().__init__(2, area, bridge, hidden)

    @property
    def available(self) -> bool:
        """Return if device is available."""
        return self._bridge.available(CONF_TEMPLATE, self._area, CONF_ROOM)

    @property
    def category(self) -> str:
        """Return the category of the entity: light, switch, or cover."""
        return "switch"

    @property
    def unique_id(self) -> str:
        """Return the ID of this room switch."""
        return "dynalite_area_" + str(self._area) + "_room_switch"

    @property
    def is_on(self) -> bool:
        """Return true if device is on."""
        device = self.get_device(1)
        assert isinstance(device, DynalitePresetSwitchDevice)
        return device.is_on

    async def async_turn_on(self, **kwargs) -> None:
        """Turn switch on."""
        # pylint: disable=unused-argument
        device = self.get_device(1)
        assert isinstance(device, DynalitePresetSwitchDevice)
        await device.async_turn_on()

    async def async_turn_off(self, **kwargs) -> None:
        """Turn switch off."""
        # pylint: disable=unused-argument
        device = self.get_device(2)
        assert isinstance(device, DynalitePresetSwitchDevice)
        await device.async_turn_on()

    def init_level(self, level):
        """Initialize to on/off."""
        on_device = self.get_device(1)
        off_device = self.get_device(2)
        if level:
            on_device.init_level(1)
            off_device.init_level(0)
        else:
            off_device.init_level(1)
            on_device.init_level(0)



"""Class to create devices from a Dynalite hub."""

import asyncio
from typing import Any, Callable, Dict, List, Optional, Set, Union

from .config import DynaliteConfig
from .const import (
    ACTIVE_INIT,
    ACTIVE_ON,
    CONF_ACT_LEVEL,
    CONF_ACTION,
    CONF_ACTION_CMD,
    CONF_ACTION_PRESET,
    CONF_ACTION_REPORT,
    CONF_ACTION_STOP,
    CONF_AREA,
    CONF_AREA_OVERRIDE,
    CONF_CHANNEL,
    CONF_CHANNEL_COVER,
    CONF_CHANNEL_TYPE,
    CONF_CLOSE_PRESET,
    CONF_DEVICE_CLASS,
    CONF_DURATION,
    CONF_FADE,
    CONF_HIDDEN_ENTITY,
    CONF_LEVEL,
    CONF_NAME,
    CONF_NONE,
    CONF_OPEN_PRESET,
    CONF_PRESET,
    CONF_QUERY_CHANNEL,
    CONF_ROOM,
    CONF_ROOM_OFF,
    CONF_ROOM_ON,
    CONF_STOP_PRESET,
    CONF_TEMPLATE,
    CONF_TILT_TIME,
    CONF_TIME_COVER,
    CONF_TRGT_LEVEL,
    DEFAULT_CHANNEL_TYPE,
    DEFAULT_COVER_CLASS,
    EVENT_CHANNEL,
    EVENT_CONNECTED,
    EVENT_DISCONNECTED,
    EVENT_PACKET,
    EVENT_PRESET,
    LOGGER,
    NOTIFICATION_PACKET,
    NOTIFICATION_PRESET,
)
from .cover import DynaliteTimeCoverDevice, DynaliteTimeCoverWithTiltDevice
from .dynalite import Dynalite
from .dynalitebase import DynaliteBaseDevice
from .event import DynetEvent
from .light import DynaliteChannelLightDevice
from .switch import (
    DynaliteChannelSwitchDevice,
    DynaliteDualPresetSwitchDevice,
    DynalitePresetSwitchDevice,
)


class DynaliteNotification:
    """A notification from the network that is sent to the application."""

    def __init__(self, notification: str, data: Dict[str, Any]):
        """Create a notification."""
        self.notification = notification
        self.data = data

    def __repr__(self):
        """Print a notification for logs."""
        return (
            "DynaliteNotification(notification="
            + self.notification
            + ", data="
            + str(self.data)
            + ")"
        )

    def __eq__(self, other):
        """Compare two notification, mostly for debug."""
        return self.notification == other.notification and self.data == other.data


class DynaliteDevices:
    """Manages a single Dynalite bridge."""

    def __init__(
        self,
        new_device_func: Callable[[List[DynaliteBaseDevice]], None],
        update_device_func: Callable[[Optional[DynaliteBaseDevice]], None],
        notification_func: Callable[[DynaliteNotification], None],
    ) -> None:
        """Initialize the system."""
        self._host = ""
        self._port = 0
        self.name = None  # public
        self._poll_timer = 0.0
        self._default_fade = 0.0
        self._default_query_channel = 0
        self._active = ""
        self._auto_discover = None
        self._loop: Optional[asyncio.AbstractEventLoop] = None
        self._new_device_func = new_device_func
        self._update_device_func = update_device_func
        self._notification_func = notification_func
        self._configured = False
        self.connected = False  # public
        self._added_presets: Dict[int, Any] = {}
        self._added_channels: Dict[int, Any] = {}
        self._added_room_switches: Dict[int, Any] = {}
        self._added_time_covers: Dict[int, Any] = {}
        self._waiting_devices: List[DynaliteBaseDevice] = []
        self._timer_active = False
        self._timer_callbacks: Set[Callable[[], None]] = set()
        self._area: Dict[int, Any] = {}
        self._dynalite = Dynalite(broadcast_func=self.handle_event)
        self._resetting = False
        self._default_presets: Dict[int, Any] = {}

    async def async_setup(self) -> bool:
        """Set up a Dynalite bridge based on host parameter in the config."""
        LOGGER.debug("bridge async_setup")
        self._loop = asyncio.get_running_loop()
        # Run the dynalite object. Assumes self.configure() has been called
        self._resetting = False
        self.connected = await self._dynalite.connect(self._host, self._port)
        return self.connected

    def configure(self, config: Dict[str, Any]) -> None:
        """Configure a Dynalite bridge."""
        LOGGER.debug("bridge async_configure - %s", config)
        self._configured = False
        configurator = DynaliteConfig(config)
        # insert the global values
        self._host = configurator.host
        self._port = configurator.port
        self.name = configurator.name
        self._auto_discover = configurator.auto_discover
        self._active = configurator.active
        self._poll_timer = configurator.poll_timer
        self._default_fade = configurator.default_fade
        self._default_query_channel = configurator.default_query_channel
        # keep the old values in case of a reconfigure, for auto discovery
        old_area = self._area
        self._area = configurator.area
        for area in old_area:
            if area not in self._area:
                self._area[area] = old_area[area]
        self._default_presets = configurator.default_presets
        # now register the channels and presets and ask for initial status if needed
        for area in self._area:
            if self._active in [ACTIVE_INIT, ACTIVE_ON]:
                self.request_area_preset(area, self._area[area][CONF_QUERY_CHANNEL])
            for channel in self._area[area][CONF_CHANNEL]:
                self.create_channel_if_new(area, channel)
                if self._active in [ACTIVE_INIT, ACTIVE_ON]:
                    self.request_channel_level(area, channel)
            for preset in self._area[area][CONF_PRESET]:
                self.create_preset_if_new(area, preset)
        # register the rooms (switches on presets 1/4)
        # all the devices should be created for channels and presets
        self.register_rooms()
        # register the time covers
        self.register_time_covers()
        # callback for all devices
        if self._new_device_func and self._waiting_devices:
            self._new_device_func(self._waiting_devices)
            self._waiting_devices = []
        self._configured = True

    def register_rooms(self) -> None:
        """Register the room switches from two normal presets each."""
        for area, area_config in self._area.items():
            if area_config.get(CONF_TEMPLATE, "") == CONF_ROOM:
                if area in self._added_room_switches:
                    continue
                new_device = DynaliteDualPresetSwitchDevice(area, self, False)
                self._added_room_switches[area] = new_device
                new_device.set_device(
                    1, self._added_presets[area][area_config[CONF_ROOM_ON]]
                )
                new_device.set_device(
                    2, self._added_presets[area][area_config[CONF_ROOM_OFF]]
                )
                self.register_new_device(new_device)

    def register_time_covers(self) -> None:
        """Register the time covers from three presets and a channel each."""
        for area, area_config in self._area.items():
            if area_config.get(CONF_TEMPLATE, "") == CONF_TIME_COVER:
                if area in self._added_time_covers:
                    continue
                if area_config[CONF_TILT_TIME] == 0:
                    new_device = DynaliteTimeCoverDevice(
                        area, self, self._poll_timer, False
                    )
                else:
                    new_device = DynaliteTimeCoverWithTiltDevice(
                        area, self, self._poll_timer, False
                    )
                self._added_time_covers[area] = new_device
                new_device.set_device(
                    1, self._added_presets[area][area_config[CONF_OPEN_PRESET]]
                )
                new_device.set_device(
                    2, self._added_presets[area][area_config[CONF_CLOSE_PRESET]]
                )
                new_device.set_device(
                    3, self._added_presets[area][area_config[CONF_STOP_PRESET]]
                )
                if area_config[CONF_CHANNEL_COVER] != 0:
                    channel_device = self._added_channels[area][
                        area_config[CONF_CHANNEL_COVER]
                    ]
                    new_device.set_device(4, channel_device)
                self.register_new_device(new_device)

    def register_new_device(self, device: DynaliteBaseDevice) -> None:
        """Register a new device and group all the ones prior to CONFIGURED event together."""
        # after initial configuration, every new device gets sent on its own. The initial ones are bunched together
        if not device.hidden:
            if self._configured:
                self._new_device_func([device])
            else:  # send all the devices together when configured
                self._waiting_devices.append(device)

    def available(self, conf: str, area: int, item_num: Union[int, str]) -> bool:
        """Return whether a device on the bridge is available."""
        if not self.connected:
            return False
        if conf in [CONF_CHANNEL, CONF_PRESET]:
            return bool(self._area.get(area, {}).get(conf, {}).get(item_num, False))
        assert conf == CONF_TEMPLATE
        return self._area.get(area, {}).get(CONF_TEMPLATE, "") == item_num

    def update_device(self, device: Optional[DynaliteBaseDevice] = None) -> None:
        """Update one or more devices."""
        if device and device.hidden:
            return
        self._update_device_func(device)

    def send_notification(self, notification: DynaliteNotification) -> None:
        """Update one or more devices."""
        self._notification_func(notification)

    def handle_event(self, event: DynetEvent) -> None:
        """Handle all events."""
        LOGGER.debug("handle_event - type=%s event=%s", event.event_type, event.data)
        if event.event_type == EVENT_CONNECTED:
            LOGGER.debug("Received CONNECTED message")
            self.connected = True
            self.update_device()
        elif event.event_type == EVENT_DISCONNECTED:
            LOGGER.debug("Received DISCONNECTED message")
            self.connected = False
            self.update_device()
        elif event.event_type == EVENT_PRESET:
            LOGGER.debug("Received PRESET message")
            assert event.data
            self.handle_preset_selection(event)
            self.send_notification(
                DynaliteNotification(
                    NOTIFICATION_PRESET,
                    {
                        CONF_AREA: event.data[CONF_AREA],
                        CONF_PRESET: event.data[CONF_PRESET],
                    },
                )
            )
        elif event.event_type == EVENT_CHANNEL:
            LOGGER.debug("Received CHANNEL message")
            self.handle_channel_change(event)
        else:
            assert event.event_type == EVENT_PACKET
            assert event.data
            LOGGER.debug("Received PACKET message")
            self.send_notification(
                DynaliteNotification(
                    NOTIFICATION_PACKET, {NOTIFICATION_PACKET: event.data[EVENT_PACKET]}
                )
            )

    def ensure_area(self, area: int) -> None:
        """Configure a default area if it is not yet in config."""
        if area not in self._area:
            LOGGER.debug("adding area %s that is not in config", area)
            # consider adding default presets to new areas (XXX)
            self._area[area] = DynaliteConfig.configure_area(
                area, {}, self._default_fade, self._default_query_channel, {}, {}
            )

    def create_preset_if_new(self, area: int, preset: int) -> None:
        """Register a new preset."""
        LOGGER.debug("create_preset_if_new - area=%s preset=%s", area, preset)
        # if already configured, ignore
        if self._added_presets.get(area, {}).get(preset, False):
            return
        self.ensure_area(area)
        area_config = self._area[area]
        if preset not in area_config[CONF_PRESET]:
            area_config[CONF_PRESET][preset] = DynaliteConfig.configure_preset(
                preset,
                self._default_presets.get(preset, {}),
                area_config[CONF_FADE],
                CONF_TEMPLATE in area_config or not self._auto_discover,
            )
            # if the area is a template is a template, new presets should be hidden
            if area_config.get(CONF_TEMPLATE, False):
                area_config[CONF_PRESET][preset][CONF_HIDDEN_ENTITY] = True
        hidden = area_config[CONF_PRESET][preset].get(CONF_HIDDEN_ENTITY, False)
        new_device = DynalitePresetSwitchDevice(area, preset, self, hidden)
        new_device.set_level(0)
        self.register_new_device(new_device)
        if area not in self._added_presets:
            self._added_presets[area] = {}
        self._added_presets[area][preset] = new_device
        LOGGER.debug(
            "Creating Dynalite preset area=%s preset=%s hidden=%s", area, preset, hidden
        )

    def handle_preset_selection(self, event: DynetEvent) -> None:
        """Change the selected preset."""
        assert event.data
        LOGGER.debug("handle_preset_selection - event=%s", event.data)
        area = event.data[CONF_AREA]
        preset = event.data[CONF_PRESET]
        self.create_preset_if_new(area, preset)
        # Update all the preset devices
        for cur_preset_in_area in self._added_presets[area]:
            device = self._added_presets[area][cur_preset_in_area]
            if cur_preset_in_area == preset:
                device.set_level(1)
            else:
                device.set_level(0)
            self.update_device(device)
        # If active is set to full, query all channels in the area
        if self._active == ACTIVE_ON:
            for channel in self._area[area].get(CONF_CHANNEL, {}):
                self.request_channel_level(area, channel)

    def create_channel_if_new(self, area: int, channel: int) -> None:
        """Register a new channel."""
        LOGGER.debug("create_channel_if_new - area=%s, channel=%s", area, channel)
        # if already configured, ignore
        if self._added_channels.get(area, {}).get(channel, False):
            return
        self.ensure_area(area)
        area_config = self._area[area]
        if channel not in area_config[CONF_CHANNEL]:
            area_config[CONF_CHANNEL][channel] = DynaliteConfig.configure_channel(
                channel,
                {},
                area_config[CONF_FADE],
                CONF_TEMPLATE in area_config or not self._auto_discover,
            )
        channel_config = area_config[CONF_CHANNEL][channel]
        LOGGER.debug("create_channel_if_new - channel_config=%s", channel_config)
        channel_type = channel_config.get(
            CONF_CHANNEL_TYPE, DEFAULT_CHANNEL_TYPE
        ).lower()
        hidden = channel_config.get(CONF_HIDDEN_ENTITY, False)
        if channel_type == "light":
            new_device: DynaliteBaseDevice = DynaliteChannelLightDevice(
                area, channel, self, hidden
            )
            self.register_new_device(new_device)
        elif channel_type == "switch":
            new_device = DynaliteChannelSwitchDevice(area, channel, self, hidden)
            self.register_new_device(new_device)
        else:
            LOGGER.info("unknown chnanel type %s - ignoring", channel_type)
            return
        if area not in self._added_channels:
            self._added_channels[area] = {}
        self._added_channels[area][channel] = new_device
        LOGGER.debug("Creating Dynalite channel area=%s channel=%s", area, channel)

    def handle_channel_change(self, event: DynetEvent) -> None:
        """Change the level of a channel."""
        assert event.data
        LOGGER.debug("handle_channel_change - data=%s", event.data)
        area = event.data[CONF_AREA]
        channel = event.data.get(CONF_CHANNEL, None)
        if channel:
            self.create_channel_if_new(area, channel)
        action = event.data[CONF_ACTION]
        if action == CONF_ACTION_REPORT:
            actual_level = (255 - event.data[CONF_ACT_LEVEL]) / 254
            target_level = (255 - event.data[CONF_TRGT_LEVEL]) / 254
            channel_to_set = self._added_channels[area][channel]
            channel_to_set.update_level(actual_level, target_level)
            self.update_device(channel_to_set)
        elif action == CONF_ACTION_CMD:
            target_level = (255 - event.data[CONF_TRGT_LEVEL]) / 254
            # when there is only a "set channel level" command, assume that this is both the actual and the target
            channel_to_set = self._added_channels[area][channel]
            channel_to_set.update_level(target_level, target_level)
            self.update_device(channel_to_set)
        elif action == CONF_ACTION_STOP:
            if channel:
                channel_to_set = self._added_channels[area][channel]
                channel_to_set.stop_fade()
                self.update_device(channel_to_set)
            else:
                for channel in self._added_channels.get(area, {}):
                    channel_to_set = self._added_channels[area][channel]
                    channel_to_set.stop_fade()
                    self.update_device(channel_to_set)
        else:
            assert action == CONF_ACTION_PRESET
            assert channel  # XXX - not handling for all channels
            area_config = self._area[area]
            area_preset = area_config.get(CONF_PRESET, {})
            preset_num = event.data[CONF_PRESET]
            target_level = area_preset.get(preset_num, {}).get(CONF_LEVEL, -1)
            if target_level != -1:
                channel_to_set = self._added_channels[area][channel]
                channel_to_set.update_level(target_level, target_level)
                self.update_device(channel_to_set)

    def add_timer_listener(self, callback_func: Callable[[], None]) -> None:
        """Add a listener to the timer and start if needed."""
        self._timer_callbacks.add(callback_func)
        if not self._timer_active:
            assert self._loop
            self._loop.call_later(self._poll_timer, self.timer_func)
            self._timer_active = True

    def remove_timer_listener(self, callback_func: Callable[[], None]) -> None:
        """Remove a listener from a timer."""
        self._timer_callbacks.discard(callback_func)

    def timer_func(self) -> None:
        """Call callbacks and either schedule timer or stop."""
        if self._timer_callbacks and not self._resetting:
            assert self._loop
            cur_callbacks = self._timer_callbacks.copy()
            for callback in cur_callbacks:
                callback()
            self._loop.call_later(self._poll_timer, self.timer_func)
        else:
            self._timer_active = False

    def set_channel_level(
        self, area: int, channel: int, level: float, fade: float
    ) -> None:
        """Set the level for a channel."""
        fade = self._area[area][CONF_CHANNEL][channel][CONF_FADE]
        self._dynalite.set_channel_level(area, channel, level, fade)

    def select_preset(self, area: int, preset: int, fade: float) -> None:
        """Select a preset in an area."""
        self._dynalite.select_preset(area, preset, fade)

    def request_area_preset(self, area: int, query_channel: Optional[int]) -> None:
        """Send a request to an area to report the preset."""
        if query_channel is None:
            if area in self._area:
                query_channel = self._area[area][CONF_QUERY_CHANNEL]
            else:
                query_channel = self._default_query_channel
        self._dynalite.request_area_preset(area, query_channel)

    def request_channel_level(self, area: int, channel: int) -> None:
        """Send a request to an area to report the preset."""
        self._dynalite.request_channel_level(area, channel)

    def get_area_name(self, area: int) -> str:
        """Return the name of an area."""
        return self._area[area][CONF_NAME]

    def get_channel_name(self, area: int, channel: int) -> str:
        """Return the name of a channel."""
        cur_area = self._area.get(area, {})
        default_area_name = f"Area {area}"
        default_channel_name = f"Channel {channel}"
        return f"{cur_area.get(CONF_NAME, default_area_name)} {cur_area.get(CONF_CHANNEL, {}).get(channel, {}).get(CONF_NAME, default_channel_name)}"

    def get_channel_fade(self, area: int, channel: int) -> float:
        """Return the fade of a channel."""
        try:
            return self._area[area][CONF_CHANNEL][channel][CONF_FADE]
        except KeyError:
            return self._default_fade

    def get_preset_name(self, area: int, preset: int) -> str:
        """Return the name of a preset."""
        cur_area = self._area.get(area, {})
        area_name = cur_area.get(CONF_NAME, f"Area {area}")
        preset_name = (
            cur_area.get(CONF_PRESET, {})
            .get(preset, {})
            .get(CONF_NAME, f"Preset {preset}")
        )
        if area_name == preset_name:
            return preset_name
        return f"{area_name} {preset_name}"

    def get_preset_fade(self, area: int, preset: int) -> float:
        """Return the fade of a preset."""
        try:
            return self._area[area][CONF_PRESET][preset][CONF_FADE]
        except KeyError:
            return self._default_fade

    def get_multi_name(self, area: int) -> str:
        """Return the name of a multi-device."""
        return self._area[area][CONF_NAME]

    def get_device_class(self, area: int) -> str:
        """Return the class for a blind."""
        try:
            return self._area[area][CONF_DEVICE_CLASS]
        except KeyError:
            return DEFAULT_COVER_CLASS

    def get_cover_duration(self, area: int) -> float:
        """Return the class for a blind."""
        try:
            return self._area[area][CONF_DURATION]
        except KeyError:
            return 60

    def get_cover_tilt_duration(self, area: int) -> float:
        """Return the class for a blind."""
        try:
            return self._area[area][CONF_TILT_TIME]
        except KeyError:
            return 0

    def get_master_area(self, area: int) -> str:
        """Get the master area when combining entities from different Dynet areas to the same area."""
        assert area in self._area
        area_config = self._area[area]
        master_area = area_config[CONF_NAME]
        if CONF_AREA_OVERRIDE in area_config:
            override_area = area_config[CONF_AREA_OVERRIDE]
            master_area = override_area if override_area.lower() != CONF_NONE else ""
        return master_area

    async def async_reset(self) -> None:
        """Reset the connections and timers."""
        self._resetting = True
        await self._dynalite.async_reset()
        while self._timer_active:
            await asyncio.sleep(0.1)



"""Support for Dynalite channels as lights."""

from typing import TYPE_CHECKING

from .const import ATTR_BRIGHTNESS
from .dynalitebase import DynaliteChannelBaseDevice

if TYPE_CHECKING:  # pragma: no cover
    from .dynalite_devices import DynaliteDevices


class DynaliteChannelLightDevice(DynaliteChannelBaseDevice):
    """Representation of a Dynalite Channel as a Home Assistant Light."""

    def __init__(
        self, area: int, channel: int, bridge: "DynaliteDevices", hidden: bool
    ) -> None:
        """Initialize the light."""
        self._level = 0.0
        self._direction = "stop"
        super().__init__(area, channel, bridge, hidden)

    @property
    def category(self) -> str:
        """Return the category of the entity: light, switch, or cover."""
        return "light"

    @property
    def brightness(self) -> int:
        """Return the brightness of this light between 0..255."""
        return int(self._level * 255)

    @property
    def level(self) -> float:
        """Return the brightness of this light between 0..255."""
        return self._level

    @property
    def direction(self) -> str:
        """Return the brightness of this light between 0..1."""
        return self._direction

    @property
    def is_on(self) -> bool:
        """Return true if device is on."""
        return self._level > 0

    def update_level(self, actual_level: float, target_level: float) -> None:
        """Update the current level."""
        old_level = self._level
        self._level = actual_level
        if target_level > actual_level:
            self._direction = "open"
        elif target_level < actual_level:
            self._direction = "close"
        else:
            self._direction = "stop"

        if self._level != old_level:
            self.update_listeners()

    async def async_turn_on(self, **kwargs) -> None:
        """Turn light on."""
        # pylint: disable=unused-argument
        if ATTR_BRIGHTNESS in kwargs:
            brightness = kwargs[ATTR_BRIGHTNESS] / 255.0
        else:
            brightness = 1.0
        fade = self._bridge.get_channel_fade(self._area, self._channel)
        self._bridge.set_channel_level(self._area, self._channel, brightness, fade)

    async def async_turn_off(self, **kwargs) -> None:
        """Turn light off."""
        # pylint: disable=unused-argument
        await self.async_turn_on(**{ATTR_BRIGHTNESS: 0})

    def init_level(self, level):
        """Initialize to a given level."""
        if level < 0 or level > 255:
            raise ValueError
        self._level = level / 255



"""Tests for DynaliteDevices."""


import pytest

import dynalite_devices_lib.const as dyn_const
from dynalite_devices_lib.dynet import DynetPacket

from .common import packet_notification, preset_notification


@pytest.mark.asyncio
async def test_empty_dynalite_devices(mock_gateway):
    """Test the dynalite devices library with no devices."""
    mock_gateway.configure_dyn_dev(
        {
            dyn_const.CONF_ACTIVE: False,
            dyn_const.CONF_AREA: {"1": {dyn_const.CONF_NO_DEFAULT: True}},
        },
        0,
    )
    assert await mock_gateway.async_setup_dyn_dev()
    await mock_gateway.check_single_update(None)
    await mock_gateway.check_writes([])


@pytest.mark.asyncio
@pytest.mark.parametrize("active", [False, dyn_const.ACTIVE_INIT, True])
async def test_dynalite_devices_active(mock_gateway, active):
    """Test with active set to ON."""
    [_, _, device_pres] = mock_gateway.configure_dyn_dev(
        {
            dyn_const.CONF_ACTIVE: active,
            dyn_const.CONF_AREA: {"1": {dyn_const.CONF_CHANNEL: {"1": {}, "2": {}}}},
            dyn_const.CONF_PRESET: {"1": {}},
        },
        3,
    )
    assert await mock_gateway.async_setup_dyn_dev()
    await mock_gateway.check_single_update(None)
    if active is not False:
        await mock_gateway.check_writes(
            [
                DynetPacket.request_channel_level_packet(1, 1),
                DynetPacket.request_channel_level_packet(1, 2),
                DynetPacket.request_area_preset_packet(1, 1),
            ]
        )
    else:
        await mock_gateway.check_writes([])
    packet_to_send = DynetPacket.report_area_preset_packet(1, 1)
    await mock_gateway.receive(packet_to_send)
    await mock_gateway.check_single_update(device_pres)
    await mock_gateway.check_notifications(
        [packet_notification(packet_to_send.raw_msg), preset_notification(1, 1)]
    )
    if active is True:
        await mock_gateway.check_writes(
            [
                DynetPacket.request_channel_level_packet(1, 1),
                DynetPacket.request_channel_level_packet(1, 2),
            ]
        )
    else:
        await mock_gateway.check_writes([])


@pytest.mark.asyncio
async def test_dynalite_devices_reconfig(mock_gateway):
    """Test reconfiguration and that no devices are registered again."""
    config = {
        dyn_const.CONF_ACTIVE: False,
        dyn_const.CONF_AREA: {
            "1": {dyn_const.CONF_CHANNEL: {"1": {}, "2": {}}},
            "2": {dyn_const.CONF_TEMPLATE: dyn_const.CONF_ROOM},
            "4": {dyn_const.CONF_TEMPLATE: dyn_const.CONF_TIME_COVER},
        },
        dyn_const.CONF_PRESET: {"1": {}},
    }
    mock_gateway.configure_dyn_dev(config, 5)
    assert await mock_gateway.async_setup_dyn_dev()
    await mock_gateway.check_single_update(None)
    mock_gateway.configure_dyn_dev(config, 0)


@pytest.mark.asyncio
async def test_dynalite_devices_auto_discover_on(mock_gateway):
    """Test autodiscover ON."""
    config = {
        dyn_const.CONF_ACTIVE: False,
        dyn_const.CONF_AUTO_DISCOVER: True,
        dyn_const.CONF_AREA: {},
    }
    mock_gateway.configure_dyn_dev(config, 0)
    assert await mock_gateway.async_setup_dyn_dev()
    await mock_gateway.check_single_update(None)
    func = mock_gateway.new_dev_func
    func.reset_mock()
    packet_to_send = DynetPacket.report_area_preset_packet(1, 1)
    await mock_gateway.receive(packet_to_send)
    await mock_gateway.check_notifications(
        [packet_notification(packet_to_send.raw_msg), preset_notification(1, 1)]
    )
    func.assert_called_once()
    devices = func.mock_calls[0][1][0]
    assert len(devices) == 1
    assert devices[0].unique_id == "dynalite_area_1_preset_1"
    await mock_gateway.check_single_update(devices[0])
    func.reset_mock()
    packet_to_send = DynetPacket.set_channel_level_packet(2, 3, 0, 0)
    await mock_gateway.receive(packet_to_send)
    await mock_gateway.check_notifications(
        [packet_notification(packet_to_send.raw_msg)]
    )
    func.assert_called_once()
    devices = func.mock_calls[0][1][0]
    assert len(devices) == 1
    assert devices[0].unique_id == "dynalite_area_2_channel_3"
    await mock_gateway.check_single_update(devices[0])


@pytest.mark.asyncio
async def test_dynalite_devices_auto_discover_off(mock_gateway):
    """Test autodiscover OFF."""
    config = {
        dyn_const.CONF_ACTIVE: False,
        dyn_const.CONF_AUTO_DISCOVER: False,
        dyn_const.CONF_AREA: {},
    }
    mock_gateway.configure_dyn_dev(config, 0)
    assert await mock_gateway.async_setup_dyn_dev()
    await mock_gateway.check_single_update(None)
    func = mock_gateway.new_dev_func
    func.reset_mock()
    packet_to_send = DynetPacket.report_area_preset_packet(1, 1)
    await mock_gateway.receive(packet_to_send)
    await mock_gateway.check_notifications(
        [packet_notification(packet_to_send.raw_msg), preset_notification(1, 1)]
    )
    func.assert_not_called()
    packet_to_send = DynetPacket.set_channel_level_packet(2, 3, 0, 0)
    await mock_gateway.receive(packet_to_send)
    await mock_gateway.check_notifications(
        [packet_notification(packet_to_send.raw_msg)]
    )
    func.assert_not_called()


@pytest.mark.asyncio
async def test_dynalite_devices_auto_discover_template(mock_gateway):
    """Test auto discover ON when running into a template that shouldn't show the device."""
    config = {
        dyn_const.CONF_ACTIVE: False,
        dyn_const.CONF_AUTO_DISCOVER: True,
        dyn_const.CONF_AREA: {
            1: {dyn_const.CONF_TEMPLATE: dyn_const.CONF_ROOM},
            2: {dyn_const.CONF_TEMPLATE: dyn_const.CONF_TIME_COVER},
        },
    }
    mock_gateway.configure_dyn_dev(config, 2)
    assert await mock_gateway.async_setup_dyn_dev()
    await mock_gateway.check_single_update(None)
    func = mock_gateway.new_dev_func
    func.reset_mock()
    packet_to_send = DynetPacket.report_area_preset_packet(1, 2)
    await mock_gateway.receive(packet_to_send)
    await mock_gateway.check_notifications(
        [packet_notification(packet_to_send.raw_msg), preset_notification(1, 2)]
    )
    func.assert_not_called()
    packet_to_send = DynetPacket.set_channel_level_packet(2, 3, 0, 0)
    await mock_gateway.receive(packet_to_send)
    await mock_gateway.check_notifications(
        [packet_notification(packet_to_send.raw_msg)]
    )
    func.assert_not_called()


@pytest.mark.asyncio
async def test_dynalite_devices_unknown_channel_type(mock_gateway):
    """Test when config has a wrong channel type."""
    mock_gateway.configure_dyn_dev(
        {
            dyn_const.CONF_ACTIVE: False,
            dyn_const.CONF_AREA: {
                "1": {
                    dyn_const.CONF_NO_DEFAULT: True,
                    dyn_const.CONF_CHANNEL: {"1": {dyn_const.CONF_CHANNEL_TYPE: "aaa"}},
                }
            },
        },
        0,
    )
    assert await mock_gateway.async_setup_dyn_dev()
    await mock_gateway.check_single_update(None)
    await mock_gateway.check_writes([])


@pytest.mark.asyncio
async def test_dynalite_devices_area_override(mock_gateway):
    """Test that area overrides work."""
    name = "aaa"
    override_name = "bbb"
    [device] = mock_gateway.configure_dyn_dev(
        {
            dyn_const.CONF_ACTIVE: False,
            dyn_const.CONF_AREA: {
                "1": {
                    dyn_const.CONF_NAME: name,
                    dyn_const.CONF_AREA_OVERRIDE: override_name,
                    dyn_const.CONF_NO_DEFAULT: True,
                    dyn_const.CONF_CHANNEL: {"1": {}},
                }
            },
        },
        1,
    )
    assert device.area_name == name
    assert device.get_master_area == override_name


@pytest.mark.asyncio
async def test_dynalite_devices_preset_collision(mock_gateway):
    """Test that a preset defined both in the defaults and in the area ."""
    name = "aaa"
    default_name = "bbb"
    devices = mock_gateway.configure_dyn_dev(
        {
            dyn_const.CONF_ACTIVE: False,
            dyn_const.CONF_AREA: {
                "1": {dyn_const.CONF_PRESET: {"1": {dyn_const.CONF_NAME: name}}},
                "2": {},
            },
            dyn_const.CONF_PRESET: {"1": {dyn_const.CONF_NAME: default_name}, "2": {}},
        },
        4,
    )
    assert devices[0].name == "Area 1 " + name
    assert devices[2].name == "Area 2 " + default_name


@pytest.mark.asyncio
async def test_dynalite_devices_reconfig_with_missing(mock_gateway):
    """Test reconfiguration with template (room / cover) devices and see that they are not available."""
    [device_room, device_cover] = mock_gateway.configure_dyn_dev(
        {
            dyn_const.CONF_ACTIVE: False,
            dyn_const.CONF_AREA: {
                "1": {dyn_const.CONF_TEMPLATE: "room"},
                "2": {dyn_const.CONF_TEMPLATE: "timecover"},
            },
            dyn_const.CONF_PRESET: {},
        },
        2,
    )
    assert await mock_gateway.async_setup_dyn_dev()
    await mock_gateway.check_single_update(None)
    assert device_room.available
    assert device_room.name == "Area 1"
    assert device_room.unique_id == "dynalite_area_1_room_switch"
    assert device_room.area_name == "Area 1"
    assert device_room.get_master_area == "Area 1"
    assert device_room.available
    assert device_cover.name == "Area 2"
    assert device_cover.unique_id == "dynalite_area_2_time_cover"
    assert device_cover.area_name == "Area 2"
    assert device_cover.get_master_area == "Area 2"
    mock_gateway.configure_dyn_dev(
        {
            dyn_const.CONF_ACTIVE: False,
            dyn_const.CONF_AREA: {"1": {}, "2": {}},
            dyn_const.CONF_PRESET: {},
        },
        0,
    )
    assert not device_room.available
    assert not device_cover.available


@pytest.mark.asyncio
async def test_dynalite_devices_default_fade(mock_gateway):
    """Test that default fade works correctly."""
    [channel_device, preset_device] = mock_gateway.configure_dyn_dev(
        {
            dyn_const.CONF_ACTIVE: False,
            dyn_const.CONF_DEFAULT: {dyn_const.CONF_FADE: 0.5},
            dyn_const.CONF_AREA: {"1": {dyn_const.CONF_CHANNEL: {"1": {}}}},
            dyn_const.CONF_PRESET: {"1": {}},
        },
        2,
    )
    assert await mock_gateway.async_setup_dyn_dev()
    await mock_gateway.check_single_update(None)
    await mock_gateway.check_writes([])
    await channel_device.async_turn_on()
    await mock_gateway.check_single_write(
        DynetPacket.set_channel_level_packet(1, 1, 1.0, 0.5)
    )
    await mock_gateway.check_single_update(channel_device)
    await preset_device.async_turn_on()
    await mock_gateway.check_single_write(
        DynetPacket.select_area_preset_packet(1, 1, 0.5)
    )
    await mock_gateway.check_notifications([preset_notification(1, 1)])
    await mock_gateway.check_single_update(preset_device)


@pytest.mark.asyncio
async def test_dynalite_devices_request_area_preset(mock_gateway):
    """Test the command to request and area preset."""
    config = {
        dyn_const.CONF_ACTIVE: False,
        dyn_const.CONF_AREA: {"1": {}, "2": {dyn_const.CONF_QUERY_CHANNEL: 6}},
        dyn_const.CONF_DEFAULT: {dyn_const.CONF_QUERY_CHANNEL: 3},
    }
    mock_gateway.configure_dyn_dev(config, 4)
    assert await mock_gateway.async_setup_dyn_dev()
    await mock_gateway.check_single_update(None)
    mock_gateway.dyn_dev.request_area_preset(1, None)
    await mock_gateway.check_single_write(DynetPacket.request_area_preset_packet(1, 3))
    mock_gateway.dyn_dev.request_area_preset(2, None)
    await mock_gateway.check_single_write(DynetPacket.request_area_preset_packet(2, 6))
    mock_gateway.dyn_dev.request_area_preset(3, None)
    await mock_gateway.check_single_write(DynetPacket.request_area_preset_packet(3, 3))
    mock_gateway.dyn_dev.request_area_preset(4, 9)
    await mock_gateway.check_single_write(DynetPacket.request_area_preset_packet(4, 9))



"""Support for the Dynalite channels as covers."""
import asyncio
from typing import TYPE_CHECKING

from .const import ATTR_POSITION, ATTR_TILT_POSITION, CONF_TEMPLATE, CONF_TIME_COVER
from .dynalitebase import DynaliteBaseDevice, DynaliteMultiDevice
from .light import DynaliteChannelLightDevice
from .switch import DynalitePresetSwitchDevice

if TYPE_CHECKING:  # pragma: no cover
    from .dynalite_devices import DynaliteDevices


class DynaliteTimeCoverDevice(DynaliteMultiDevice):
    """Representation of a Dynalite Channel as a Home Assistant Cover."""

    def __init__(
        self, area: int, bridge: "DynaliteDevices", poll_timer: float, hidden: bool
    ) -> None:
        """Initialize the cover."""
        self._current_position = 0.0
        self._initialized = False
        self._direction = "stop"
        self._poll_timer = poll_timer
        super().__init__(4, area, bridge, hidden)

    @property
    def available(self) -> bool:
        """Return if device is available."""
        return self._bridge.available(CONF_TEMPLATE, self._area, CONF_TIME_COVER)

    @property
    def category(self) -> str:
        """Return the category of the entity: light, switch, or cover."""
        return "cover"

    @property
    def unique_id(self) -> str:
        """Return the ID of this room switch."""
        return "dynalite_area_" + str(self._area) + "_time_cover"

    @property
    def has_tilt(self) -> bool:
        """Return whether cover supports tilt."""
        return False

    @property
    def device_class(self) -> str:
        """Return the class of the cover."""
        return self._bridge.get_device_class(self._area)

    def update_level(self, actual_level: float, target_level: float) -> None:
        """Update the current level."""
        if actual_level == target_level:
            if not self._initialized:
                self._current_position = target_level
                self._initialized = True
            self._direction = "stop"
            self._bridge.remove_timer_listener(self.timer_callback)
        elif target_level > actual_level:
            if not self._initialized:
                self._current_position = 1.0
                self._initialized = True
            self._direction = "open"
            self._bridge.add_timer_listener(self.timer_callback)
        else:  # target_level < actual_level
            if not self._initialized:
                self._current_position = 0.0
                self._initialized = True
            self._direction = "close"
            self._bridge.add_timer_listener(self.timer_callback)
        self._bridge.update_device(self)

    def timer_callback(self) -> None:
        """Update the progress of open and close."""
        duration = self._bridge.get_cover_duration(self._area)
        assert self._direction in ["open", "close"]
        if self._direction == "open":
            self._current_position += self._poll_timer / duration
            getattr(self, "update_tilt", int)(self._poll_timer)
            if self._current_position >= 1.0:
                self._current_position = 1.0
                self._direction = "stop"
                self._bridge.remove_timer_listener(self.timer_callback)
        else:
            assert self._direction == "close"
            self._current_position -= self._poll_timer / duration
            getattr(self, "update_tilt", int)(self._poll_timer)
            if self._current_position <= 0.0:
                self._current_position = 0.0
                self._direction = "stop"
                self._bridge.remove_timer_listener(self.timer_callback)
        self._bridge.update_device(self)

    @property
    def current_cover_position(self) -> int:
        """Return the position of the cover from 0 to 100."""
        return int(self._current_position * 100)

    @property
    def is_opening(self) -> bool:
        """Return whether cover is currently opening."""
        return self._direction == "open"

    @property
    def is_closing(self) -> bool:
        """Return whether cover is currently closing."""
        return self._direction == "close"

    @property
    def is_closed(self) -> bool:
        """Return whether cover is closed."""
        return self._current_position == 0

    async def async_open_cover(self, **kwargs) -> None:
        """Open the cover."""
        # pylint: disable=unused-argument
        device = self.get_device(1)
        assert isinstance(device, DynalitePresetSwitchDevice)
        await device.async_turn_on()
        self.update_level(self._current_position, 1.0)

    async def async_close_cover(self, **kwargs) -> None:
        """Close the cover."""
        # pylint: disable=unused-argument
        device = self.get_device(2)
        assert isinstance(device, DynalitePresetSwitchDevice)
        await device.async_turn_on()
        self.update_level(self._current_position, 0.0)

    async def async_set_cover_position(self, **kwargs) -> None:
        """Set the cover to a specific position."""
        target_position = kwargs[ATTR_POSITION] / 100
        position_diff = target_position - self._current_position
        if position_diff > 0.001:
            await self.async_open_cover()
            while (
                self._current_position < target_position and self._direction == "open"
            ):
                await asyncio.sleep(self._poll_timer)
            if self._direction == "open":
                await self.async_stop_cover()
                await asyncio.sleep(self._poll_timer)  # doing twice for safety
                await self.async_stop_cover()
        elif position_diff < -0.001:
            await self.async_close_cover()
            while (
                self._current_position > target_position and self._direction == "close"
            ):
                await asyncio.sleep(self._poll_timer)
            if self._direction == "close":
                await self.async_stop_cover()
                await asyncio.sleep(self._poll_timer)  # doing twice for safety
                await self.async_stop_cover()
        else:
            await self.async_stop_cover()

    async def async_stop_cover(self, **kwargs) -> None:
        """Stop the cover."""
        # pylint: disable=unused-argument
        device = self.get_device(3)
        assert isinstance(device, DynalitePresetSwitchDevice)
        await device.async_turn_on()
        self.update_level(self._current_position, self._current_position)

    def listener(self, device: DynaliteBaseDevice, stop_fade: bool) -> None:
        """Update according to updates in underlying devices."""
        if device == self.get_device(1):
            assert isinstance(device, DynalitePresetSwitchDevice)
            if device.is_on:
                self.update_level(self._current_position, 1.0)
        elif device == self.get_device(2):
            assert isinstance(device, DynalitePresetSwitchDevice)
            if device.is_on:
                self.update_level(self._current_position, 0.0)
        elif device == self.get_device(3):
            assert isinstance(device, DynalitePresetSwitchDevice)
            if device.is_on:
                self.update_level(self._current_position, self._current_position)
        else:
            assert device == self.get_device(4)
            assert isinstance(device, DynaliteChannelLightDevice)
            if stop_fade or device.direction == "stop":
                self.update_level(device.level, device.level)
            elif device.direction == "open":
                self.update_level(self._current_position, 1.0)
            else:
                self.update_level(self._current_position, 0.0)
        super().listener(device, stop_fade)

    def init_level(self, level):
        """Initialize to a given position."""
        if level < 0 or level > 100:
            raise ValueError
        self._current_position = level / 100.0
        self._initialized = True


class DynaliteTimeCoverWithTiltDevice(DynaliteTimeCoverDevice):
    """Representation of a Dynalite Channel as a Home Assistant Cover that uses up and down for tilt."""

    def __init__(
        self, area: int, bridge: "DynaliteDevices", poll_timer: float, hidden: bool
    ) -> None:
        """Initialize the cover."""
        self._current_tilt = 0.0
        super().__init__(area, bridge, poll_timer, hidden)

    @property
    def has_tilt(self) -> bool:
        """Return whether cover supports tilt."""
        return True

    def update_tilt(self, poll_timer: float) -> None:
        """Update the current tilt based on diff and tilt_percentage."""
        assert self._direction in ["open", "close"]
        if self._direction == "open":
            mult = poll_timer
        else:
            assert self._direction == "close"
            mult = -poll_timer
        tilt_duration = self._bridge.get_cover_tilt_duration(self._area)
        tilt_diff = mult / tilt_duration
        self._current_tilt = max(0, min(1, self._current_tilt + tilt_diff))

    @property
    def current_cover_tilt_position(self) -> int:
        """Return the current cover tilt."""
        return int(self._current_tilt * 100)

    async def apply_tilt_diff(self, tilt_diff: float) -> None:
        """Move the cover up or down based on a diff."""
        duration = self._bridge.get_cover_duration(self._area)
        tilt_duration = self._bridge.get_cover_tilt_duration(self._area)
        factor = tilt_duration / duration
        position_diff = tilt_diff * factor
        target_position = int(
            100 * max(0, min(1, self._current_position + position_diff))
        )
        await self.async_set_cover_position(position=target_position)

    async def async_open_cover_tilt(self, **kwargs) -> None:
        """Open the cover tilt."""
        # pylint: disable=unused-argument
        if self._current_tilt == 1:
            return
        await self.apply_tilt_diff(1 - self._current_tilt)

    async def async_close_cover_tilt(self, **kwargs) -> None:
        """Close the cover tilt."""
        # pylint: disable=unused-argument
        if self._current_tilt == 0:
            return
        await self.apply_tilt_diff(0 - self._current_tilt)

    async def async_set_cover_tilt_position(self, **kwargs) -> None:
        """Set the cover tilt position."""
        target_position = kwargs[ATTR_TILT_POSITION] / 100
        await self.apply_tilt_diff(target_position - self._current_tilt)

    async def async_stop_cover_tilt(self, **kwargs) -> None:
        """Stop cover tilt."""
        # pylint: disable=unused-argument
        await self.async_stop_cover()



"""
Class that handles inbound requests on the Dynet network and fires events.

@ Author      : Troy Kelly
@ Date        : 23 Sept 2018
@ Description : Philips Dynalite Library - Unofficial interface for Philips Dynalite over RS485

@ Notes:        Requires a RS485 to IP gateway (Do not use the Dynalite one - use something cheaper)
"""

from .const import (
    CONF_ACT_LEVEL,
    CONF_ACTION,
    CONF_ACTION_CMD,
    CONF_ACTION_PRESET,
    CONF_ACTION_REPORT,
    CONF_ACTION_STOP,
    CONF_AREA,
    CONF_CHANNEL,
    CONF_FADE,
    CONF_PRESET,
    CONF_TRGT_LEVEL,
    EVENT_CHANNEL,
    EVENT_PRESET,
)
from .dynet import DynetPacket
from .event import DynetEvent


class DynetInbound:
    """Class to handle inboud Dynet packets."""

    def __init__(self) -> None:
        """Initialize the object."""
        pass

    @staticmethod
    def preset(packet: DynetPacket) -> DynetEvent:
        """Handle a preset that was selected."""
        if packet.command > 3:
            preset = packet.command - 6
        else:
            preset = packet.command
        preset = (preset + (packet.data[2] * 8)) + 1
        fade = (packet.data[0] + (packet.data[1] * 256)) * 0.02
        return DynetEvent(
            event_type=EVENT_PRESET,
            data={CONF_AREA: packet.area, CONF_PRESET: preset, CONF_FADE: fade},
        )

    def preset_1(self, packet: DynetPacket) -> DynetEvent:
        """Handle preset 1 in banks of 8."""
        return self.preset(packet)

    def preset_2(self, packet: DynetPacket) -> DynetEvent:
        """Handle preset 2 in banks of 8."""
        return self.preset(packet)

    def preset_3(self, packet: DynetPacket) -> DynetEvent:
        """Handle preset 3 in banks of 8."""
        return self.preset(packet)

    def preset_4(self, packet: DynetPacket) -> DynetEvent:
        """Handle preset 4 in banks of 8."""
        return self.preset(packet)

    def preset_5(self, packet: DynetPacket) -> DynetEvent:
        """Handle preset 5 in banks of 8."""
        return self.preset(packet)

    def preset_6(self, packet: DynetPacket) -> DynetEvent:
        """Handle preset 6 in banks of 8."""
        return self.preset(packet)

    def preset_7(self, packet: DynetPacket) -> DynetEvent:
        """Handle preset 7 in banks of 8."""
        return self.preset(packet)

    def preset_8(self, packet: DynetPacket) -> DynetEvent:
        """Handle preset 8 in banks of 8."""
        return self.preset(packet)

    @staticmethod
    def report_preset(packet: DynetPacket) -> DynetEvent:
        """Report the current preset of an area."""
        preset = packet.data[0] + 1
        return DynetEvent(
            event_type=EVENT_PRESET, data={CONF_AREA: packet.area, CONF_PRESET: preset},
        )

    @staticmethod
    def linear_preset(packet: DynetPacket) -> DynetEvent:
        """Report that preset was selected with fade."""
        preset = packet.data[0] + 1
        fade = (packet.data[1] + (packet.data[2] * 256)) * 0.02
        return DynetEvent(
            event_type=EVENT_PRESET,
            data={CONF_AREA: packet.area, CONF_PRESET: preset, CONF_FADE: fade},
        )

    @staticmethod
    def report_channel_level(packet: DynetPacket) -> DynetEvent:
        """Report the new level of a channel."""
        channel = packet.data[0] + 1
        target_level = packet.data[1]
        actual_level = packet.data[2]
        return DynetEvent(
            event_type=EVENT_CHANNEL,
            data={
                CONF_AREA: packet.area,
                CONF_CHANNEL: channel,
                CONF_ACTION: CONF_ACTION_REPORT,
                CONF_TRGT_LEVEL: target_level,
                CONF_ACT_LEVEL: actual_level,
            },
        )

    @staticmethod
    def set_channel_x_to_level_with_fade(
        packet: DynetPacket, channel_offset: int
    ) -> DynetEvent:
        """Report that a channel was set to a specific level."""
        channel = ((packet.data[1] + 1) % 256) * 4 + channel_offset
        target_level = packet.data[0]
        return DynetEvent(
            event_type=EVENT_CHANNEL,
            data={
                CONF_AREA: packet.area,
                CONF_CHANNEL: channel,
                CONF_ACTION: CONF_ACTION_CMD,
                CONF_TRGT_LEVEL: target_level,
            },
        )

    def set_channel_1_to_level_with_fade(self, packet: DynetPacket) -> DynetEvent:
        """Report that channel 1 was set to a specific level."""
        return self.set_channel_x_to_level_with_fade(packet, 1)

    def set_channel_2_to_level_with_fade(self, packet: DynetPacket) -> DynetEvent:
        """Report that channel 2 was set to a specific level."""
        return self.set_channel_x_to_level_with_fade(packet, 2)

    def set_channel_3_to_level_with_fade(self, packet: DynetPacket) -> DynetEvent:
        """Report that channel 3 was set to a specific level."""
        return self.set_channel_x_to_level_with_fade(packet, 3)

    def set_channel_4_to_level_with_fade(self, packet: DynetPacket) -> DynetEvent:
        """Report that channel 4 was set to a specific level."""
        return self.set_channel_x_to_level_with_fade(packet, 4)

    @staticmethod
    def request_channel_level(packet: DynetPacket) -> None:
        """Do nothing."""
        pass

    @staticmethod
    def stop_fading(packet: DynetPacket) -> DynetEvent:
        """Report that fading stopped for a channel or area."""
        data = {CONF_AREA: packet.area, CONF_ACTION: CONF_ACTION_STOP}
        channel = packet.data[0] + 1
        if channel != 256:  # all channels in area
            data[CONF_CHANNEL] = channel
        return DynetEvent(event_type=EVENT_CHANNEL, data=data)

    @staticmethod
    def fade_channel_area_to_preset(packet: DynetPacket) -> DynetEvent:
        """Report that fading stopped for a channel or area XXX."""
        fade = packet.data[2] * 0.02
        data = {
            CONF_AREA: packet.area,
            CONF_ACTION: CONF_ACTION_PRESET,
            CONF_PRESET: packet.data[1] + 1,
            CONF_FADE: fade,
        }
        channel = packet.data[0] + 1
        if channel != 256:  # all channels in area
            data[CONF_CHANNEL] = channel
        return DynetEvent(event_type=EVENT_CHANNEL, data=data)



"""Tests for Dynalite coverss."""

import asyncio

import pytest

import dynalite_devices_lib.const as dyn_const
from dynalite_devices_lib.dynet import DynetPacket

from .common import packet_notification, preset_notification


@pytest.mark.asyncio
async def test_cover_no_tilt(mock_gateway):
    """Test basic cover functionality."""
    name = "NAME"
    [
        channel_device,
        open_device,
        close_device,
        stop_device,
        cover_device,
    ] = mock_gateway.configure_dyn_dev(
        {
            dyn_const.CONF_ACTIVE: False,
            dyn_const.CONF_POLL_TIMER: 0.05,
            dyn_const.CONF_AREA: {
                "1": {
                    dyn_const.CONF_NAME: name,
                    dyn_const.CONF_TEMPLATE: dyn_const.CONF_TIME_COVER,
                    dyn_const.CONF_DURATION: 0.5,
                    dyn_const.CONF_OPEN_PRESET: 1,
                    dyn_const.CONF_CLOSE_PRESET: 2,
                    dyn_const.CONF_STOP_PRESET: 3,
                    dyn_const.CONF_CHANNEL_COVER: 4,
                    dyn_const.CONF_CHANNEL: {"4": {}},
                    dyn_const.CONF_PRESET: {"1": {}, "2": {}, "3": {}},
                }
            },
        },
        5,
    )
    assert await mock_gateway.async_setup_dyn_dev()
    await mock_gateway.check_single_update(None)
    assert channel_device.category == "light"
    assert open_device.category == "switch"
    assert close_device.category == "switch"
    assert stop_device.category == "switch"
    assert cover_device.category == "cover"
    assert cover_device.name == name
    assert cover_device.unique_id == "dynalite_area_1_time_cover"
    assert cover_device.available
    assert cover_device.area_name == name
    assert cover_device.get_master_area == name
    assert not cover_device.has_tilt
    assert cover_device.device_class == dyn_const.DEFAULT_COVER_CLASS
    # Initialize to closed - otherwise it takes the first one as the init status
    packet_to_send = DynetPacket.report_area_preset_packet(1, 2)
    await mock_gateway.receive(packet_to_send)
    await mock_gateway.check_updates(
        [open_device, close_device, stop_device, cover_device], True
    )
    await mock_gateway.check_notifications(
        [packet_notification(packet_to_send.raw_msg), preset_notification(1, 2)]
    )
    # It is closed. Let's open
    assert cover_device.is_closed
    await cover_device.async_open_cover()
    await mock_gateway.check_updates(
        [open_device, close_device, stop_device, cover_device], True
    )
    await mock_gateway.check_single_write(
        DynetPacket.select_area_preset_packet(1, 1, 0)
    )
    await mock_gateway.check_notifications([preset_notification(1, 1)])
    assert open_device.is_on
    assert not close_device.is_on
    assert not stop_device.is_on
    await asyncio.sleep(0.25)
    assert cover_device.is_opening
    assert 40 < cover_device.current_cover_position < 60
    await asyncio.sleep(0.4)
    assert (
        not cover_device.is_closed
        and not cover_device.is_opening
        and not cover_device.is_closing
    )
    assert cover_device.current_cover_position == 100
    await mock_gateway.check_updates([cover_device], True)
    # It is open. Now let's close
    await cover_device.async_close_cover()
    await mock_gateway.check_updates(
        [open_device, close_device, stop_device, cover_device], True
    )
    await mock_gateway.check_single_write(
        DynetPacket.select_area_preset_packet(1, 2, 0)
    )
    await mock_gateway.check_notifications([preset_notification(1, 2)])
    assert close_device.is_on
    assert not open_device.is_on
    assert not stop_device.is_on
    await asyncio.sleep(0.25)
    await mock_gateway.check_updates([cover_device], True)
    assert cover_device.is_closing
    assert 40 < cover_device.current_cover_position < 60
    # Stop halfway
    await cover_device.async_stop_cover()
    await mock_gateway.check_updates(
        [open_device, close_device, stop_device, cover_device], True
    )
    await mock_gateway.check_single_write(
        DynetPacket.select_area_preset_packet(1, 3, 0)
    )
    await mock_gateway.check_notifications([preset_notification(1, 3)])
    assert stop_device.is_on
    assert not open_device.is_on
    assert not close_device.is_on
    assert (
        not cover_device.is_closed
        and not cover_device.is_opening
        and not cover_device.is_closing
    )
    # And continue to full close
    await cover_device.async_close_cover()
    await mock_gateway.check_updates(
        [open_device, close_device, stop_device, cover_device], True
    )
    await mock_gateway.check_single_write(
        DynetPacket.select_area_preset_packet(1, 2, 0)
    )
    await mock_gateway.check_notifications([preset_notification(1, 2)])
    await asyncio.sleep(0.4)
    await mock_gateway.check_updates([cover_device], True)
    assert cover_device.is_closed
    assert cover_device.current_cover_position == 0
    # Now open it half-way
    await cover_device.async_set_cover_position(position=50)
    await mock_gateway.check_updates(
        [open_device, close_device, stop_device, cover_device], True
    )
    await mock_gateway.check_notifications(
        [
            preset_notification(1, 1),
            preset_notification(1, 3),
            preset_notification(1, 3),
        ]
    )
    assert (
        not cover_device.is_closed
        and not cover_device.is_opening
        and not cover_device.is_closing
    )
    assert 40 < cover_device.current_cover_position < 60
    await cover_device.async_set_cover_position(position=25)
    await mock_gateway.check_updates(
        [open_device, close_device, stop_device, cover_device], True
    )
    await mock_gateway.check_notifications(
        [
            preset_notification(1, 2),
            preset_notification(1, 3),
            preset_notification(1, 3),
        ]
    )
    assert (
        not cover_device.is_closed
        and not cover_device.is_opening
        and not cover_device.is_closing
    )
    assert 15 < cover_device.current_cover_position < 35
    await cover_device.async_open_cover()
    await mock_gateway.check_updates(
        [open_device, close_device, stop_device, cover_device], True
    )
    await mock_gateway.check_notifications([preset_notification(1, 1)])
    assert cover_device.is_opening
    mock_gateway.reset()
    await cover_device.async_set_cover_position(
        position=cover_device.current_cover_position
    )
    await mock_gateway.check_updates(
        [open_device, close_device, stop_device, cover_device], True
    )
    await mock_gateway.check_notifications([preset_notification(1, 3)])
    await mock_gateway.check_single_write(
        DynetPacket.select_area_preset_packet(1, 3, 0)
    )
    assert (
        not cover_device.is_closed
        and not cover_device.is_opening
        and not cover_device.is_closing
    )
    # Now send commands
    packet_to_send = DynetPacket.report_area_preset_packet(1, 1)
    await mock_gateway.receive(packet_to_send)
    await mock_gateway.check_updates(
        [open_device, close_device, stop_device, cover_device], True
    )
    await mock_gateway.check_notifications(
        [packet_notification(packet_to_send.raw_msg), preset_notification(1, 1)]
    )
    assert cover_device.is_opening
    packet_to_send = DynetPacket.report_area_preset_packet(1, 2)
    await mock_gateway.receive(packet_to_send)
    await mock_gateway.check_updates(
        [open_device, close_device, stop_device, cover_device], True
    )
    await mock_gateway.check_notifications(
        [packet_notification(packet_to_send.raw_msg), preset_notification(1, 2)]
    )
    assert cover_device.is_closing
    packet_to_send = DynetPacket.report_area_preset_packet(1, 3)
    await mock_gateway.receive(packet_to_send)
    await mock_gateway.check_updates(
        [open_device, close_device, stop_device, cover_device], True
    )
    await mock_gateway.check_notifications(
        [packet_notification(packet_to_send.raw_msg), preset_notification(1, 3)]
    )
    assert not cover_device.is_closing and not cover_device.is_opening
    packet_to_send = DynetPacket.report_area_preset_packet(1, 1)
    await mock_gateway.receive(packet_to_send)
    await mock_gateway.check_updates(
        [open_device, close_device, stop_device, cover_device], True
    )
    await mock_gateway.check_notifications(
        [packet_notification(packet_to_send.raw_msg), preset_notification(1, 1)]
    )
    await asyncio.sleep(0.3)
    await mock_gateway.check_updates([cover_device], True)
    packet_to_send = DynetPacket.report_channel_level_packet(1, 4, 0, 1)
    await mock_gateway.receive(packet_to_send)
    await mock_gateway.check_updates([cover_device, channel_device], True)
    await mock_gateway.check_notifications(
        [packet_notification(packet_to_send.raw_msg)]
    )
    assert cover_device.is_closing
    await asyncio.sleep(0.01)
    # Open and then stop with channel
    packet_to_send = DynetPacket.report_channel_level_packet(1, 4, 1, 0)
    await mock_gateway.receive(packet_to_send)
    await mock_gateway.check_updates([cover_device, channel_device], True)
    await mock_gateway.check_notifications(
        [packet_notification(packet_to_send.raw_msg)]
    )
    assert cover_device.is_opening
    packet_to_send = DynetPacket.stop_channel_fade_packet(1, 4)
    await mock_gateway.receive(packet_to_send)
    await mock_gateway.check_updates([cover_device, channel_device], True)
    await mock_gateway.check_notifications(
        [packet_notification(packet_to_send.raw_msg)]
    )
    assert not cover_device.is_opening
    assert not cover_device.is_closing
    # Open and then stop with area (channel=0xff)
    packet_to_send = DynetPacket.report_channel_level_packet(1, 4, 0, 1)
    await mock_gateway.receive(packet_to_send)
    await mock_gateway.check_updates([cover_device, channel_device], True)
    await mock_gateway.check_notifications(
        [packet_notification(packet_to_send.raw_msg)]
    )
    await asyncio.sleep(0.1)
    packet_to_send = DynetPacket.report_channel_level_packet(1, 4, 1, 0)
    await mock_gateway.receive(packet_to_send)
    await mock_gateway.check_updates([cover_device, channel_device], True)
    await mock_gateway.check_notifications(
        [packet_notification(packet_to_send.raw_msg)]
    )
    assert cover_device.is_opening
    packet_to_send = DynetPacket.stop_channel_fade_packet(1, 256)
    await mock_gateway.receive(packet_to_send)
    await mock_gateway.check_updates([cover_device, channel_device], True)
    await mock_gateway.check_notifications(
        [packet_notification(packet_to_send.raw_msg)]
    )
    assert not cover_device.is_opening
    assert not cover_device.is_closing
    cover_device.init_level(50)
    assert cover_device.current_cover_position == 50
    cover_device.init_level(0)
    assert cover_device.is_closed
    with pytest.raises(ValueError):
        cover_device.init_level(-1)
    with pytest.raises(ValueError):
        cover_device.init_level(101)


@pytest.mark.asyncio
async def test_cover_with_tilt(mock_gateway):
    """Test the cover tilt functionality."""
    name = "NAME"
    [cover_device] = mock_gateway.configure_dyn_dev(
        {
            dyn_const.CONF_ACTIVE: False,
            dyn_const.CONF_POLL_TIMER: 0.05,
            dyn_const.CONF_AREA: {
                "1": {
                    dyn_const.CONF_NAME: name,
                    dyn_const.CONF_TEMPLATE: dyn_const.CONF_TIME_COVER,
                    dyn_const.CONF_DEVICE_CLASS: "blind",
                    dyn_const.CONF_DURATION: 0.5,
                    dyn_const.CONF_TILT_TIME: 0.25,
                    dyn_const.CONF_OPEN_PRESET: 1,
                    dyn_const.CONF_CLOSE_PRESET: 2,
                    dyn_const.CONF_STOP_PRESET: 3,
                    dyn_const.CONF_CHANNEL_COVER: 4,
                }
            },
        }
    )
    assert await mock_gateway.async_setup_dyn_dev()
    await mock_gateway.check_single_update(None)
    assert cover_device.category == "cover"
    assert cover_device.device_class == "blind"
    assert cover_device.has_tilt
    # Initialize to closed - otherwise it takes the first one as the init status
    packet_to_send = DynetPacket.report_area_preset_packet(1, 2)
    await mock_gateway.receive(packet_to_send)
    await mock_gateway.check_updates([cover_device], True)
    await mock_gateway.check_notifications(
        [packet_notification(packet_to_send.raw_msg), preset_notification(1, 2)]
    )
    # It is closed. Let's open
    assert cover_device.is_closed
    assert cover_device.current_cover_tilt_position == 0
    asyncio.create_task(cover_device.async_open_cover_tilt())
    await mock_gateway.check_single_write(
        DynetPacket.select_area_preset_packet(1, 1, 0)
    )
    await mock_gateway.check_updates([cover_device], True)
    await mock_gateway.check_notifications([preset_notification(1, 1)])
    await asyncio.sleep(0.125)
    await mock_gateway.check_updates([cover_device], True)
    assert cover_device.is_opening
    assert 30 < cover_device.current_cover_tilt_position < 70
    # Stop halfway
    await cover_device.async_stop_cover_tilt()
    await mock_gateway.check_updates([cover_device], True)
    await mock_gateway.check_single_write(
        DynetPacket.select_area_preset_packet(1, 3, 0)
    )
    await mock_gateway.check_notifications([preset_notification(1, 3)])
    await cover_device.async_open_cover_tilt()
    await mock_gateway.check_notifications(
        [
            preset_notification(1, 1),
            preset_notification(1, 3),
            preset_notification(1, 3),
        ]
    )
    await mock_gateway.check_updates([cover_device], True)
    assert cover_device.current_cover_tilt_position == 100
    assert 30 < cover_device.current_cover_position < 70
    assert (
        not cover_device.is_closed
        and not cover_device.is_opening
        and not cover_device.is_closing
    )
    await asyncio.sleep(0.1)
    mock_gateway.reset()
    asyncio.create_task(cover_device.async_open_cover_tilt())
    await mock_gateway.check_writes([])
    # It is open. Now let's close
    mock_gateway.reset()
    asyncio.create_task(cover_device.async_close_cover_tilt())
    await mock_gateway.check_updates([cover_device], True)
    await mock_gateway.check_single_write(
        DynetPacket.select_area_preset_packet(1, 2, 0)
    )
    await mock_gateway.check_notifications([preset_notification(1, 2)])
    await asyncio.sleep(0.125)
    await mock_gateway.check_updates([cover_device], True)
    assert cover_device.is_closing
    assert 30 < cover_device.current_cover_tilt_position < 70
    # Stop halfway
    await cover_device.async_stop_cover_tilt()
    await mock_gateway.check_updates([cover_device], True)
    await mock_gateway.check_single_write(
        DynetPacket.select_area_preset_packet(1, 3, 0)
    )
    await mock_gateway.check_notifications([preset_notification(1, 3)])
    assert 30 < cover_device.current_cover_tilt_position < 70
    assert (
        not cover_device.is_closed
        and not cover_device.is_opening
        and not cover_device.is_closing
    )
    # And continue to full close
    await cover_device.async_close_cover()
    await mock_gateway.check_updates([cover_device], True)
    await mock_gateway.check_single_write(
        DynetPacket.select_area_preset_packet(1, 2, 0)
    )
    await mock_gateway.check_notifications([preset_notification(1, 2)])
    await asyncio.sleep(0.6)
    await mock_gateway.check_updates([cover_device], True)
    assert cover_device.is_closed
    assert cover_device.current_cover_tilt_position == 0
    mock_gateway.reset()
    asyncio.create_task(cover_device.async_close_cover_tilt())
    await mock_gateway.check_updates([])
    await mock_gateway.check_writes([])
    # Now open it half-way
    await cover_device.async_set_cover_tilt_position(tilt_position=50)
    await mock_gateway.check_notifications(
        [
            preset_notification(1, 1),
            preset_notification(1, 3),
            preset_notification(1, 3),
        ]
    )
    await mock_gateway.check_updates([cover_device], True)
    await asyncio.sleep(0.3)
    await mock_gateway.check_updates([])
    assert (
        not cover_device.is_closed
        and not cover_device.is_opening
        and not cover_device.is_closing
    )
    assert 30 < cover_device.current_cover_tilt_position < 70


@pytest.mark.asyncio
async def test_cover_no_channel(mock_gateway):
    """Test a cover without an associated channel."""
    name = "NAME"
    [cover_device] = mock_gateway.configure_dyn_dev(
        {
            dyn_const.CONF_ACTIVE: False,
            dyn_const.CONF_POLL_TIMER: 0.05,
            dyn_const.CONF_AREA: {
                "1": {
                    dyn_const.CONF_NAME: name,
                    dyn_const.CONF_TEMPLATE: dyn_const.CONF_TIME_COVER,
                    dyn_const.CONF_DURATION: 0.5,
                    dyn_const.CONF_OPEN_PRESET: 1,
                    dyn_const.CONF_CLOSE_PRESET: 2,
                    dyn_const.CONF_STOP_PRESET: 3,
                }
            },
        }
    )
    assert await mock_gateway.async_setup_dyn_dev()
    await mock_gateway.check_single_update(None)
    assert cover_device.category == "cover"
    # Initialize to closed - otherwise it takes the first one as the init status
    packet_to_send = DynetPacket.report_area_preset_packet(1, 2)
    await mock_gateway.receive(packet_to_send)
    await mock_gateway.check_updates([cover_device], True)
    await mock_gateway.check_notifications(
        [packet_notification(packet_to_send.raw_msg), preset_notification(1, 2)]
    )
    # It is closed. Let's open
    assert cover_device.is_closed
    await cover_device.async_open_cover()
    await mock_gateway.check_single_write(
        DynetPacket.select_area_preset_packet(1, 1, 0)
    )
    await mock_gateway.check_updates([cover_device], True)
    await mock_gateway.check_notifications([preset_notification(1, 1)])
    await asyncio.sleep(0.25)
    await mock_gateway.check_updates([cover_device], True)
    assert cover_device.is_opening
    assert 40 < cover_device.current_cover_position < 60
    await asyncio.sleep(0.4)
    await mock_gateway.check_updates([cover_device], True)
    assert (
        not cover_device.is_closed
        and not cover_device.is_opening
        and not cover_device.is_closing
    )
    assert cover_device.current_cover_position == 100
